"""View model for MVVM demo of DataSelector."""

from typing import Any, Dict

from nova.mvvm.interface import BindingInterface
from tests.gallery.models.neutron_data_selector import (
    AnalysisDataSelectorState,
    NeutronDataSelectorBindingTest,
    ONCatDataSelectorState,
)


class NeutronDataSelectorVM:
    """View model for MVVM demo of NeutronDataSelector."""

    def __init__(self, binding: BindingInterface) -> None:
        self.analysis_model = AnalysisDataSelectorState()
        self.oncat_model = ONCatDataSelectorState()
        self.binding_test = NeutronDataSelectorBindingTest()

        self.analysis_bind = binding.new_bind(self.analysis_model, callback_after_update=self.on_update)
        self.oncat_bind = binding.new_bind(self.oncat_model)
        self.parameter_bind = binding.new_bind(self.binding_test, callback_after_update=self.on_parameter_update)

    def on_update(self, data: Dict[str, Any]) -> None:
        print(data, self.analysis_model)

    def on_parameter_update(self, data: Dict[str, Any]) -> None:
        print(data, self.binding_test)



from .data_selector import DataSelector
from .execution_buttons import ExecutionButtons
from .file_upload import FileUpload
from .input_field import InputField
from .progress_bar import ProgressBar
from .remote_file_input import RemoteFileInput
from .tool_outputs import ToolOutputWindows

__all__ = [
    "DataSelector",
    "ExecutionButtons",
    "FileUpload",
    "InputField",
    "ProgressBar",
    "RemoteFileInput",
    "ToolOutputWindows",
]



"""Utilities for our layout components."""

from typing import Dict, Union


def merge_styles(*styles: Union[Dict[str, str], str]) -> str:
    result = ""

    for style in styles:
        if isinstance(style, dict):
            for key, value in style.items():
                result += f" {key}: {value};"
        else:
            result += f" {style};"

    return result



"""Sphinx configuration."""

import os
import sys

import tomli

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "NOVA Application Development"
copyright = "2025, ORNL"
author = "John Duggan"
with open("../pyproject.toml", "rb") as toml_file:
    toml_dict = tomli.load(toml_file)
    release = toml_dict["project"]["version"]

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

sys.path.insert(0, os.path.abspath("../src/nova/trame"))

exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
extensions = ["sphinx.ext.autodoc", "sphinx.ext.napoleon", "sphinx_rtd_theme"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"



"""Unit tests for DataSelector."""

from typing import List

from pydantic import BaseModel, Field
from trame.app import get_server
from trame_server.core import Server

from nova.mvvm.trame_binding import TrameBinding
from nova.trame.view.components import DataSelector
from nova.trame.view.theme import ThemedApp


def test_data_selector() -> None:
    class MyTrameApp(ThemedApp):
        def __init__(self, server: Server = None) -> None:
            server = get_server(None, client_type="vue3")
            super().__init__(server=server)
            self.create_ui()

        def create_ui(self) -> None:
            with super().create_ui() as layout:
                with layout.content:
                    input = DataSelector(v_model="test", directory="/", extensions=[".tiff"])
                    assert input.v_model == "test"
                    assert input._directory == "/"
                    assert input._extensions == [".tiff"]

    MyTrameApp()


def test_parameter_bindings() -> None:
    class TestModel(BaseModel):
        v_model: List[str] = Field(default=[])
        directory: str = Field(default="")
        extensions: List[str] = Field(default=[])
        refresh_rate: int = Field(default=0)
        select_strategy: str = Field(default="")
        subdirectory: str = Field(default="")

    class MyTrameApp(ThemedApp):
        def __init__(self, server: Server = None) -> None:
            self.server = get_server(None, client_type="vue3")
            super().__init__(server=self.server)

            self.create_binding()
            self.create_ui()

        def create_binding(self) -> None:
            self.test_obj = TestModel()

            binding = TrameBinding(self.server.state)
            self.test_binding = binding.new_bind(self.test_obj)
            self.test_binding.connect("test_ds")

        def create_ui(self) -> None:
            with super().create_ui() as layout:
                with layout.content:
                    DataSelector(
                        v_model=("test_ds.v_model", ["test.txt"]),
                        directory=("test_ds.directory", "/"),
                        extensions=("test_ds.extensions", [".txt"]),
                        subdirectory=("test_ds.subdirectory", "bin"),
                        refresh_rate=("test_ds.refresh_rate", 15),
                        select_strategy=("test_ds.select_strategy", "page"),
                    )

    MyTrameApp()



"""Model for MVVM demo of DataSelector."""

from typing import List

from pydantic import BaseModel, Field


class DataSelectorBindingTest(BaseModel):
    """Model for testing binding parameters for DataSelector."""

    directory: str = Field(default="", title="Base Directory")
    refresh_rate: int = Field(default=0, title="Refresh Rate")
    subdirectory: str = Field(default="")


class DataSelectorState(BaseModel):
    """Model for MVVM demo of DataSelector."""

    selected_files: List[str] = Field(default=[], title="Selected Files")



"""Unit tests for InputField."""

from nova.trame.view.components import RemoteFileInput


def test_remote_file_input() -> None:
    file_input = RemoteFileInput(v_model="test")
    assert file_input.allow_files
    assert not file_input.allow_folders


def test_open_dialog() -> None:
    file_input = RemoteFileInput(v_model="test")
    file_input.vm.open_dialog()
    assert file_input.vm.previous_value == ""
    assert file_input.vm.value == ""


def test_close_dialog() -> None:
    file_input = RemoteFileInput(v_model="test")
    file_input.vm.open_dialog()

    file_input.vm.set_value("test")
    file_input.vm.close_dialog(cancel=True)
    assert file_input.vm.previous_value == ""
    assert file_input.vm.value == file_input.vm.previous_value

    file_input.vm.set_value("test")
    file_input.vm.close_dialog()
    assert file_input.vm.previous_value == ""
    assert file_input.vm.value == "test"


def test_toggle_show_all() -> None:
    file_input = RemoteFileInput(v_model="test")
    assert not file_input.vm.showing_all_files
    file_input.vm.toggle_showing_all_files()
    assert file_input.vm.showing_all_files
    file_input.vm.toggle_showing_all_files()
    assert not file_input.vm.showing_all_files


def test_select_file() -> None:
    file_input = RemoteFileInput(v_model="test")
    file_input.vm.select_file("")
    assert file_input.vm.value == ""
    file_input.vm.select_file("/")
    assert file_input.vm.value == "/"
    assert not file_input.vm.model.valid_selection("/")
    file_input.vm.select_file("usr")
    assert file_input.vm.value == "/usr"
    file_input.vm.select_file({"path": "local"})
    assert file_input.vm.value == "/usr/local"
    file_input.vm.select_file("..")
    file_input.vm.select_file("..")
    file_input.vm.select_file("..")
    assert file_input.vm.value == ""
    file_input.vm.select_file("bogus_path")
    assert file_input.vm.value == "bogus_path"



"""View implementation for FileUpload."""

from typing import Any, Iterable, Tuple, Union

from trame.app import get_server
from trame.widgets import vuetify3 as vuetify
from trame_server.core import State

from nova.trame._internal.utils import get_state_param
from nova.trame.utils.types import TrameTuple

from .remote_file_input import RemoteFileInput


class FileUpload(vuetify.VBtn):
    """Component for uploading a file from either the user's filesystem or the server filesystem."""

    def __init__(
        self,
        v_model: Union[str, Tuple],
        base_paths: Union[Iterable[str], str, None] = None,
        extensions: Union[Iterable[str], str, None] = None,
        label: str = "",
        return_contents: Union[bool, Tuple] = True,
        show_server_files: Union[bool, Tuple] = True,
        use_bytes: Union[bool, Tuple] = False,
        **kwargs: Any,
    ) -> None:
        """Constructor for FileUpload.

        For all parameters, tuples have a special syntax. See :ref:`TrameTuple <api_trame_tuple>` for a description of
        it.

        Parameters
        ----------
        v_model : Union[str, Tuple]
            The state variable to set when the user uploads their file. The state variable will contain a latin1-decoded
            version of the file contents. If your file is binary or requires a different string encoding, then you can
            call `encode('latin1')` on the file contents to get the underlying bytes.
        base_paths: Union[Iterable[str], str], optional
            Passed to :ref:`RemoteFileInput <api_remotefileinput>`. Typical Trame binding syntax doesn't work here as
            tuples are interpreted as literal extensions to filter with. Instead, you can pass a string with a
            JavaScript expression to bind this parameter.
        extensions: Union[Iterable[str], str], optional
            Restricts the files shown to the user to files that end with one of the strings in the list. Typical Trame
            binding syntax doesn't work here as tuples are interpreted as literal extensions to filter with. Instead,
            you can pass a string with a JavaScript expression to bind this parameter.
        label : str, optional
            The text to display on the upload button.
        return_contents : Union[bool, Tuple], optional
            If true, the file contents will be stored in v_model. If false, a file path will be stored in v_model.
            Defaults to true.
        show_server_files : Union[bool, Tuple], optional
            If true, then the "From Server" option to select a file will be shown. Defaults to true.
        use_bytes : Union[bool, Tuple], optional
            If true, then files uploaded from the local machine will contain bytes rather than text.
        **kwargs
            All other arguments will be passed to the underlying
            `Button component <https://trame.readthedocs.io/en/latest/trame.widgets.vuetify3.html#trame.widgets.vuetify3.VBtn>`_.

        Returns
        -------
        None
        """
        self._server = get_server(None, client_type="vue3")

        self._v_model = v_model
        self._base_paths = base_paths if base_paths else ["/"]
        self._extensions = extensions if extensions else []
        self._return_contents = return_contents
        self._use_bytes = use_bytes
        self._show_server_files = TrameTuple.create(show_server_files)
        self._ref_name = f"nova__fileupload_{self._next_id}"

        super().__init__(
            label, click=f"!{self._show_server_files.expression} && trame.refs.{self._ref_name}.click()", **kwargs
        )
        self.create_ui()

    @property
    def state(self) -> State:
        return self._server.state

    def create_ui(self) -> None:
        self.local_file_input = vuetify.VFileInput(
            v_model=self._v_model,
            __properties=["accept"],
            accept=",".join(
                get_state_param(self.state, (self._extensions,))
                if isinstance(self._extensions, str)
                else self._extensions
            ),
            classes="d-none",
            ref=self._ref_name,
            # Serialize the content in a way that will work with nova-mvvm and then push it to the server.
            update_modelValue=(
                f"{self._v_model}.arrayBuffer().then((contents) => {{"
                f"  trigger('decode_blob_{self._id}', [contents]); "
                "});"
            ),
        )
        self.remote_file_input = RemoteFileInput(
            v_model=self._v_model,
            base_paths=self._base_paths,
            extensions=self._extensions,
            input_props={"classes": "d-none"},
            return_contents=self._return_contents,
            use_bytes=self._use_bytes,
        )

        with self:
            with vuetify.VMenu(v_if=self._show_server_files.expression, activator="parent"):
                with vuetify.VList():
                    vuetify.VListItem("From Local Machine", click=f"trame.refs.{self._ref_name}.click()")
                    vuetify.VListItem("From Server", click=self.remote_file_input.open_dialog)

        @self.server.controller.trigger(f"decode_blob_{self._id}")
        def _decode_blob(contents: bytes) -> None:
            self.remote_file_input.decode_file(contents, get_state_param(self.state, self._return_contents))

    def select_file(self, value: str) -> None:
        """Programmatically set the RemoteFileInput path.

        Parameters
        ----------
        value: str
            The new value for the RemoteFileInput.

        Returns
        -------
        None
        """
        self.remote_file_input.select_file(value)



"""pytest Fixture for setting up the Selenium driver and the server."""

from functools import partial
from multiprocessing import Process
from time import sleep
from typing import Generator

from pytest import fixture
from selenium.webdriver import Firefox, FirefoxOptions

from tests.gallery import main


def _setup_selenium() -> Firefox:
    options = FirefoxOptions()
    options.add_argument("-headless")
    options.add_argument("--height=20000")

    driver = Firefox(options=options)
    driver.get("http://localhost:8080")
    driver.implicitly_wait(10)

    return driver


@fixture(autouse=True, scope="session")
def driver() -> Generator[Firefox, None, None]:
    server_process = Process(target=partial(main, open_browser=False))
    server_process.start()

    # TODO: Surely there's a better way to wait for the server to start.
    sleep(1)

    yield _setup_selenium()

    server_process.terminate()



"""Model implementation for DataSelector."""

import os
from pathlib import Path
from typing import Any, Dict, List, Optional

from natsort import natsorted
from pydantic import BaseModel, Field


class DataSelectorState(BaseModel, validate_assignment=True):
    """Selection state for identifying datafiles."""

    directory: str = Field(default="")
    extensions: List[str] = Field(default=[])
    search: str = Field(default="", title="Search")
    subdirectory: str = Field(default="")


class DataSelectorModel:
    """Manages file system interactions for the DataSelector widget."""

    def __init__(self, state: DataSelectorState) -> None:
        self.state: DataSelectorState = state

    def set_binding_parameters(self, **kwargs: Any) -> None:
        if "directory" in kwargs:
            self.state.directory = kwargs["directory"]
        if "extensions" in kwargs:
            self.state.extensions = kwargs["extensions"]
        if "subdirectory" in kwargs:
            self.state.subdirectory = kwargs["subdirectory"]

    def sort_directories(self, directories: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        # Sort the current level of dictionaries
        sorted_dirs = natsorted(directories, key=lambda x: x["title"])

        # Process each sorted item to sort their children
        for item in sorted_dirs:
            if "children" in item and isinstance(item["children"], list):
                item["children"] = self.sort_directories(item["children"])

        return sorted_dirs

    def get_directories_from_path(self, base_path: Path) -> List[Dict[str, Any]]:
        directories = []
        try:
            for dirpath, dirs, _ in os.walk(base_path):
                # Get the relative path from the start path
                path_parts = os.path.relpath(dirpath, base_path).split(os.sep)

                if len(path_parts) > 1:
                    dirs.clear()
                elif path_parts != ["."]:
                    # Subdirectories are fully queried upon being opened, so we only need to query one item to determine
                    # if the target directory has any children.
                    dirs[:] = dirs[:1]

                # Only create a new entry for top-level directories
                if len(path_parts) == 1 and path_parts[0] != ".":  # This indicates a top-level directory
                    current_dir = {"path": dirpath, "title": path_parts[0]}
                    directories.append(current_dir)

                # Add subdirectories to the corresponding parent directory
                elif len(path_parts) > 1:
                    current_level: Any = directories
                    for part in path_parts[:-1]:  # Parent directories
                        for item in current_level:
                            if item["title"] == part:
                                if "children" not in item:
                                    item["children"] = []
                                current_level = item["children"]
                                break

                    # Add the last part (current directory) as a child
                    current_level.append({"path": dirpath, "title": path_parts[-1]})
        except OSError:
            pass

        return self.sort_directories(directories)

    def get_directories(self, base_path: Optional[Path] = None) -> List[Dict[str, Any]]:
        if base_path:
            pass
        else:
            base_path = Path(self.state.directory)

        if not base_path:
            return []

        return self.get_directories_from_path(base_path)

    def get_datafiles_from_path(self, base_path: Path) -> List[str]:
        datafiles = []
        try:
            datafile_path = base_path / self.state.subdirectory

            for entry in os.scandir(datafile_path):
                can_add = False
                if entry.is_file():
                    if self.state.extensions:
                        for extension in self.state.extensions:
                            if entry.path.lower().endswith(extension):
                                can_add = True
                                break
                    else:
                        can_add = True

                if self.state.search and self.state.search.lower() not in entry.name.lower():
                    can_add = False

                if can_add:
                    datafiles.append(entry.path)
        except OSError:
            pass

        return natsorted(datafiles)

    def get_datafiles(self) -> List[Dict[str, str]]:
        base_path = Path(self.state.directory)

        return [{"path": datafile} for datafile in self.get_datafiles_from_path(base_path)]

    def set_subdirectory(self, subdirectory_path: str) -> None:
        self.state.subdirectory = subdirectory_path



"""View model implementation for the DataSelector widget."""

import os
from pathlib import Path
from typing import Any, Dict, List

from nova.mvvm.interface import BindingInterface
from nova.trame.model.data_selector import DataSelectorModel


class DataSelectorViewModel:
    """Manages the view state of the DataSelector widget."""

    def __init__(self, model: DataSelectorModel, binding: BindingInterface) -> None:
        self.model: DataSelectorModel = model

        self.datafiles: List[Dict[str, Any]] = []
        self.directories: List[Dict[str, Any]] = []
        self.expanded: Dict[str, List[str]] = {}

        self.state_bind = binding.new_bind(self.model.state, callback_after_update=self.on_state_updated)
        self.directories_bind = binding.new_bind()
        self.datafiles_bind = binding.new_bind()
        self.reset_bind = binding.new_bind()

    def expand_directory(self, paths: List[str]) -> None:
        if paths[-1] in self.expanded:
            return

        # Query for the new subdirectories to display in the view
        new_directories = self.model.get_directories(Path(paths[-1]))

        # Find the entry in the existing directories that corresponds to the directory to expand
        current_level: Dict[str, Any] = {}
        children: List[Dict[str, Any]] = self.directories
        for current_path in paths:
            if current_level:
                children = current_level["children"]

            for entry in children:
                if current_path == entry["path"]:
                    current_level = entry
                    break
        current_level["children"] = new_directories

        # Mark this directory as expanded and display the new content
        self.expanded[paths[-1]] = paths
        self.directories_bind.update_in_view(self.directories)

    def reexpand_directories(self) -> None:
        paths_to_expand = self.expanded.values()
        self.expanded = {}

        for paths in paths_to_expand:
            self.expand_directory(paths)

    def on_state_updated(self, results: Dict[str, Any]) -> None:
        for result in results.get("updated", []):
            if result == "search":
                self.update_view()

    def set_binding_parameters(self, **kwargs: Any) -> None:
        self.model.set_binding_parameters(**kwargs)
        self.update_view(refresh_directories=True)

    def set_subdirectory(self, subdirectory_path: str = "") -> None:
        self.model.set_subdirectory(subdirectory_path)
        self.update_view()

    def transform_datafiles(self, datafiles: List[Any]) -> List[Dict[str, str]]:
        for datafile in datafiles:
            datafile["title"] = os.path.basename(datafile["path"])
        return datafiles

    def update_view(self, refresh_directories: bool = False) -> None:
        self.state_bind.update_in_view(self.model.state)
        if not self.directories or refresh_directories:
            self.directories = self.model.get_directories()
            self.reexpand_directories()
        self.directories_bind.update_in_view(self.directories)

        self.datafiles = self.transform_datafiles(self.model.get_datafiles())
        self.datafiles_bind.update_in_view(self.datafiles)



"""View Implementation for DataSelector."""

from typing import Any, List, Literal, Tuple, Union
from warnings import warn

from trame.app import get_server
from trame.widgets import vuetify3 as vuetify

from nova.mvvm._internal.utils import rgetdictvalue
from nova.mvvm.trame_binding import TrameBinding
from nova.trame.model.ornl.analysis_data_selector import (
    CUSTOM_DIRECTORIES_LABEL,
    AnalysisDataSelectorModel,
    AnalysisDataSelectorState,
)
from nova.trame.model.ornl.neutron_data_selector import NeutronDataSelectorModel
from nova.trame.model.ornl.oncat_data_selector import ONCatDataSelectorModel, ONCatDataSelectorState
from nova.trame.utils.types import TrameTuple
from nova.trame.view.layouts import GridLayout
from nova.trame.view_model.ornl.neutron_data_selector import NeutronDataSelectorViewModel

from ..data_selector import DataSelector, get_state_param, set_state_param
from ..input_field import InputField

vuetify.enable_lab()


class NeutronDataSelector(DataSelector):
    """Allows the user to select datafiles from an IPTS experiment."""

    def __init__(
        self,
        v_model: Union[str, Tuple],
        allow_custom_directories: Union[bool, Tuple] = False,
        clear_selection_on_experiment_change: Union[bool, Tuple] = True,
        data_source: Literal["filesystem", "oncat"] = "filesystem",
        facility: Union[str, Tuple] = "",
        instrument: Union[str, Tuple] = "",
        experiment: Union[str, Tuple] = "",
        show_experiment_filters: Union[bool, Tuple] = True,
        show_selected_files: Union[bool, Tuple] = True,
        extensions: Union[List[str], Tuple, None] = None,
        projection: Union[List[str], Tuple, None] = None,
        subdirectory: Union[str, Tuple] = "",
        refresh_rate: Union[int, Tuple] = 30,
        select_strategy: Union[str, Tuple] = "all",
        **kwargs: Any,
    ) -> None:
        """Constructor for DataSelector.

        For all parameters, tuples have a special syntax. See :ref:`TrameTuple <api_trame_tuple>` for a description of
        it.

        Parameters
        ----------
        v_model : Union[str, Tuple]
            The name of the state variable to bind to this widget. The state variable will contain a list of the files
            selected by the user.
        allow_custom_directories : Union[bool, Tuple], optional
            Whether or not to allow users to provide their own directories to search for datafiles in. Ignored if the
            facility parameter is set.
        clear_selection_on_experiment_change: Union[bool, Tuple], optional
            Whether or not to clear the selected files when the user changes the facility, instrument, or experiment.
        data_source : Literal["filesystem", "oncat"], optional
            The source from which to pull datafiles. Defaults to "filesystem". If using ONCat, you will need to set the
            following environment variables for local development: `ONCAT_CLIENT_ID` and `ONCAT_CLIENT_SECRET`. Note
            that this parameter does not supporting Trame bindings. If you need to swap between the options, please
            create two instances of this class and switch between them using a v_if or a v_show.
        facility : Union[str, Tuple], optional
            The facility to restrict data selection to. Options: HFIR, SNS
        instrument : Union[str, Tuple], optional
            The instrument to restrict data selection to. Please use the instrument acronym (e.g. CG-2).
        experiment : Union[str, Tuple], optional
            The experiment to restrict data selection to.
        show_experiment_filters : Union[bool, Tuple], optional
            If false, then the facility, instrument, and experiment selection widgets will be hidden from the user. This
            is only intended to be used when all of these parameters are fixed or controlled by external widgets.
        show_selected_files : Union[bool, Tuple], optional
            If true, then the currently selected files will be shown to the user below the directory and file selection
            widgets.
        extensions : Union[List[str], Tuple], optional
            A list of file extensions to restrict selection to. If unset, then all files will be shown.
        projection : Union[List[str], Tuple], optional
            Sets the projection argument when pulling data files via pyoncat. Please refer to the ONCat documentation
            for how to use this. This should only be used with `data_source="oncat"`.
        subdirectory : Union[str, Tuple], optional
            A subdirectory within the user's chosen experiment to show files. If not specified as a string, the user
            will be shown a folder browser and will be able to see all files in the experiment that they have access to.
        refresh_rate : Union[str, Tuple], optional
            The number of seconds between attempts to automatically refresh the file list. Set to zero to disable this
            feature. Defaults to 30 seconds.
        select_strategy : Union[str, Tuple], optional
            The selection strategy to pass to the `VDataTable component <https://trame.readthedocs.io/en/latest/trame.widgets.vuetify3.html#trame.widgets.vuetify3.VDataTable>`__.
            If unset, the `all` strategy will be used.
        **kwargs
            All other arguments will be passed to the underlying
            `VDataTable component <https://trame.readthedocs.io/en/latest/trame.widgets.vuetify3.html#trame.widgets.vuetify3.VDataTable>`_.

        Returns
        -------
        None
        """
        if data_source == "oncat" and allow_custom_directories:
            warn("allow_custom_directories will be ignored since data will be pulled from ONCat.", stacklevel=1)

        if data_source == "oncat" and subdirectory:
            warn("subdirectory will be ignored since data will be pulled from ONCat.", stacklevel=1)

        if isinstance(facility, str) and facility and allow_custom_directories:
            warn("allow_custom_directories will be ignored since the facility parameter is fixed.", stacklevel=1)

        self._facility = facility
        self._instrument = instrument
        self._experiment = experiment
        self._allow_custom_directories = allow_custom_directories
        self._last_allow_custom_directories = self._allow_custom_directories
        self._data_source = data_source
        self._projection = projection

        self._show_experiment_filters = TrameTuple.create(show_experiment_filters)

        self._state_name = f"nova__dataselector_{self._next_id}_state"
        self._facilities_name = f"nova__neutrondataselector_{self._next_id}_facilities"
        self._selected_facility_name = (
            self._facility[0] if isinstance(self._facility, tuple) else f"{self._state_name}.facility"
        )
        self._instruments_name = f"nova__neutrondataselector_{self._next_id}_instruments"
        self._selected_instrument_name = (
            self._instrument[0] if isinstance(self._instrument, tuple) else f"{self._state_name}.instrument"
        )
        self._experiments_name = f"nova__neutrondataselector_{self._next_id}_experiments"
        self._selected_experiment_name = (
            self._experiment[0] if isinstance(self._experiment, tuple) else f"{self._state_name}.experiment"
        )

        super().__init__(
            v_model,
            "",
            clear_selection_on_directory_change=clear_selection_on_experiment_change,
            extensions=extensions,
            subdirectory=subdirectory if data_source == "filesystem" else "oncat",
            refresh_rate=refresh_rate,
            select_strategy=select_strategy,
            show_selected_files=show_selected_files,
            **kwargs,
        )

    def create_projection_column_title(self, key: str) -> str:
        return key.split(".")[-1].replace("_", " ").title()

    def create_ui(self, **kwargs: Any) -> None:
        headers = kwargs.pop(
            "headers",
            None,
        )
        if self._projection and headers is None:
            headers = "[{ title: 'Available Datafiles', key: 'title' },"
            for key in self._projection:
                headers += f"{{title: '{self.create_projection_column_title(key)}', key: '{key}'}},"
            headers += "]"
        else:
            headers = "[{ title: 'Available Datafiles', key: 'title' }]"

        super().create_ui(headers=(headers,), **kwargs)

        with self._layout.filter:
            with GridLayout(v_if=self._show_experiment_filters.expression, columns=3):
                column_span = 3
                if isinstance(self._facility, tuple) or not self._facility:
                    column_span -= 1
                    InputField(
                        v_model=self._selected_facility_name,
                        items=(self._facilities_name,),
                        type="autocomplete",
                        variant="outlined",
                        update_modelValue=(self.update_facility, "[$event]"),
                    )
                if isinstance(self._instrument, tuple) or not self._instrument:
                    column_span -= 1
                    with InputField(
                        v_if=f"{self._selected_facility_name} !== '{CUSTOM_DIRECTORIES_LABEL}'",
                        v_model=self._selected_instrument_name,
                        chips=True,
                        items=(self._instruments_name,),
                        item_value="name",
                        type="autocomplete",
                        variant="outlined",
                        update_modelValue=(self.update_instrument, "[$event]"),
                    ):
                        with vuetify.Template(v_slot_chip="data"):
                            vuetify.VChip("{{ data.item.raw.id }}", v_if="data.item.raw", classes="mr-1")
                            vuetify.VListItemTitle("{{ data.item.raw.name }}")
                        with vuetify.Template(v_slot_item="data"):
                            with vuetify.VListItem(v_bind="data.props"):
                                with vuetify.Template(v_slot_prepend=True):
                                    vuetify.VChip("{{ data.item.raw.id }}", classes="mr-1")
                                with vuetify.Template(v_slot_title=True):
                                    vuetify.VListItemTitle("{{ data.item.raw.name }}")
                InputField(
                    v_if=f"{self._selected_facility_name} !== '{CUSTOM_DIRECTORIES_LABEL}'",
                    v_model=self._selected_experiment_name,
                    column_span=column_span,
                    items=(self._experiments_name,),
                    item_value="title",
                    type="autocomplete",
                    variant="outlined",
                    update_modelValue=(self.update_experiment, "[$event]"),
                )
                InputField(
                    v_else=True, v_model=f"{self._state_name}.custom_directory", column_span=2, variant="outlined"
                )

    def _create_model(self) -> None:
        self._model: NeutronDataSelectorModel
        if self._data_source == "oncat":
            self._model = ONCatDataSelectorModel(ONCatDataSelectorState())
        else:
            self._model = AnalysisDataSelectorModel(AnalysisDataSelectorState())

    def _create_viewmodel(self) -> None:
        server = get_server(None, client_type="vue3")
        binding = TrameBinding(server.state)

        self._vm: NeutronDataSelectorViewModel = NeutronDataSelectorViewModel(self._model, binding)
        self._vm.state_bind.connect(self._state_name)
        self._vm.facilities_bind.connect(self._facilities_name)
        self._vm.instruments_bind.connect(self._instruments_name)
        self._vm.experiments_bind.connect(self._experiments_name)
        self._vm.directories_bind.connect(self._directories_name)
        self._vm.datafiles_bind.connect(self._datafiles_name)
        self._vm.reset_bind.connect(self.reset)

        self._vm.update_view()

    # This method sets up Trame state change listeners for each binding parameter that can be changed directly by this
    # component. This allows us to communicate the changes to the developer's bindings without requiring our own. We
    # don't want bindings in the internal implementation as our callbacks could compete with the developer's.
    def _setup_bindings(self) -> None:
        # If the bindings were given initial values, write these to the state.
        set_state_param(self.state, self._facility)
        set_state_param(self.state, self._instrument)
        set_state_param(self.state, self._experiment)
        set_state_param(self.state, self._allow_custom_directories)
        set_state_param(self.state, self._projection)
        self._last_facility = get_state_param(self.state, self._facility)
        self._last_instrument = get_state_param(self.state, self._instrument)
        self._last_experiment = get_state_param(self.state, self._experiment)
        self._vm.set_binding_parameters(
            facility=get_state_param(self.state, self._facility),
            instrument=get_state_param(self.state, self._instrument),
            experiment=get_state_param(self.state, self._experiment),
            allow_custom_directories=get_state_param(self.state, self._allow_custom_directories),
            projection=get_state_param(self.state, self._projection),
        )

        # Now we set up the change listeners for all bound parameters. These are responsible for updating the component
        # when other portions of the application manipulate these parameters.
        if isinstance(self._facility, tuple):

            @self.state.change(self._facility[0].split(".")[0])
            def on_facility_change(**kwargs: Any) -> None:
                facility = rgetdictvalue(kwargs, self._facility[0])
                if facility != self._last_facility:
                    self._last_facility = facility
                    self._vm.set_binding_parameters(
                        facility=set_state_param(self.state, (self._selected_facility_name,), facility)
                    )
                    self._vm.reset()

        if isinstance(self._instrument, tuple):

            @self.state.change(self._instrument[0].split(".")[0])
            def on_instrument_change(**kwargs: Any) -> None:
                instrument = rgetdictvalue(kwargs, self._instrument[0])
                if instrument != self._last_instrument:
                    self._last_instrument = instrument
                    self._vm.set_binding_parameters(
                        instrument=set_state_param(self.state, (self._selected_instrument_name,), instrument)
                    )
                    self._vm.reset()

        if isinstance(self._experiment, tuple):

            @self.state.change(self._experiment[0].split(".")[0])
            def on_experiment_change(**kwargs: Any) -> None:
                experiment = rgetdictvalue(kwargs, self._experiment[0])
                if experiment and experiment != self._last_experiment:
                    self._last_experiment = experiment
                    # See the note in the update_experiment method for why we call this twice.
                    self._vm.set_binding_parameters(
                        experiment=set_state_param(self.state, (self._selected_experiment_name,), ""),
                    )
                    self._vm.set_binding_parameters(
                        experiment=set_state_param(self.state, (self._selected_experiment_name,), experiment)
                    )
                    self._vm.reset()

        if isinstance(self._allow_custom_directories, tuple):

            @self.state.change(self._allow_custom_directories[0].split(".")[0])
            def on_allow_custom_directories_change(**kwargs: Any) -> None:
                allow_custom_directories = rgetdictvalue(kwargs, self._allow_custom_directories[0])  # type: ignore
                if allow_custom_directories != self._last_allow_custom_directories:
                    self._last_allow_custom_directories = allow_custom_directories
                    self._vm.set_binding_parameters(
                        allow_custom_directories=set_state_param(
                            self.state, self._allow_custom_directories, allow_custom_directories
                        )
                    )

        if isinstance(self._projection, tuple):

            @self.state.change(self._projection[0].split(".")[0])
            def on_projection_change(**kwargs: Any) -> None:
                projection = rgetdictvalue(kwargs, self._projection[0])  # type: ignore
                if projection != self._projection:
                    self._projection = projection
                    self._vm.set_binding_parameters(
                        projection=set_state_param(self.state, self._projection, projection)
                    )

        super()._setup_bindings()

    # These update methods notify the rest of the application when the component changes bound parameters.
    def update_facility(self, facility: str) -> None:
        self._vm.set_binding_parameters(
            facility=set_state_param(self.state, (self._selected_facility_name,), facility),
            instrument=set_state_param(self.state, (self._selected_instrument_name,), ""),  # Reset the instrument
            experiment=set_state_param(self.state, (self._selected_experiment_name,), ""),  # Reset the experiment
        )
        self._vm.reset()

    def update_instrument(self, instrument: str) -> None:
        self._vm.set_binding_parameters(
            instrument=set_state_param(self.state, (self._selected_instrument_name,), instrument),
            experiment=set_state_param(self.state, (self._selected_experiment_name,), ""),  # Reset the experiment
        )
        self._vm.reset()

    def update_experiment(self, experiment: str) -> None:
        # Setting the experiment to an empty string forces the treeview to clear it's selection state.
        self._vm.set_binding_parameters(
            experiment=set_state_param(self.state, (self._selected_experiment_name,), ""),
        )
        self._vm.set_binding_parameters(
            experiment=set_state_param(self.state, (self._selected_experiment_name,), experiment),
        )
        self._vm.reset()

    def set_state(self, *args: Any, **kwargs: Any) -> None:
        raise TypeError(
            "The set_state method has been removed. Please use update_facility, update_instrument, and "
            "update_experiment instead."
        )



"""View model for MVVM demo of FileUpload."""

import sys
from typing import Any, Dict

from nova.mvvm.interface import BindingInterface
from tests.gallery.models.file_upload import FileUploadState


class FileUploadVM:
    """View model for MVVM demo of FileUpload."""

    def __init__(self, binding: BindingInterface) -> None:
        self.model = FileUploadState()
        self.model_bind = binding.new_bind(self.model, callback_after_update=self.on_update)

    def on_update(self, data: Dict[str, Any]) -> None:
        print("file size:", sys.getsizeof(self.model.file))

    def update_view(self) -> None:
        self.model_bind.update_in_view(self.model)



"""Model for MVVM demo of FileUpload."""

from typing import List

from pydantic import BaseModel, Field


class FileUploadState(BaseModel):
    """Model for MVVM demo of FileUpload."""

    extensions: List[str] = Field(default=[".cif", ".h5", ".nxs"])
    file: str = Field(default="")
    label: str = Field(default="Upload File")



"""View implementation for Interactive2DPlot."""

from typing import Any, Optional

from altair import Chart
from trame.widgets import client, vega


class Interactive2DPlot(vega.Figure):
    """Creates an interactive 2D plot in Trame using Vega.

    Trame provides two primary mechanisms for composing 2D plots: `Plotly <https://github.com/Kitware/trame-plotly>`_
    and `Vega-Lite/Altair <https://github.com/Kitware/trame-vega>`_. If you only need static plots or basic browser
    event handling, then please use these libraries directly.

    If you need to capture complex front-end interactions, then you can
    use our provided Interactive2DPlot widget that is based on Vega-Lite. This uses the same API as Trame's vega.Figure,
    except that it will automatically sync Vega's signal states as the user interacts with the plot.

    The following allows the user to select a region of the plot and tracks the selected region in Python:

    .. literalinclude:: ../tests/test_interactive_2d_plot.py
        :start-after: setup 2d plot
        :end-before: setup 2d plot complete
        :dedent:
    """

    def __init__(self, figure: Optional[Chart] = None, **kwargs: Any) -> None:
        """Constructor for Interactive2DPlot.

        Parameters
        ----------
        figure : `altair.Chart <https://altair-viz.github.io/user_guide/generated/toplevel/altair.Chart.html#altair.Chart>`__, optional
            Altair chart object
        kwargs
            Arguments to be passed to `AbstractElement <https://trame.readthedocs.io/en/latest/core.widget.html#trame_client.widgets.core.AbstractElement>`_

        Returns
        -------
        None
        """  # noqa: E501
        self._initialized = False

        super().__init__(figure=figure, **kwargs)
        self.ref = f"nova__vega_{self._id}"
        self.server.state[self.ref] = {}
        self._start_update_handlers = client.JSEval(
            exec=(
                "async () => {"
                f" let ref = window.trame.refs['{self.ref}'];"
                "  await ref.mountVis();"  # wait for the new  visualization to be rendered in the front-end
                "  if (ref.viz === undefined) { return; }"  # If the component is not mounted, do nothing
                "  for (const [key, value] of Object.entries(ref.viz.view._signals)) {"
                "    if (key === 'unit') { continue; }"  # this causes a JSError for some reason if not skipped
                "    ref.viz.view.addSignalListener(key, (name, value) => {"
                f"     window.trame.state.state['{self.ref}'][name] = value;"  # sync front-end state
                f"     flushState('{self.ref}');"  # sync back-end state
                "    })"
                "  }"
                "}"
            )
        ).exec

        client.ClientTriggers(mounted=self.update)

    def get_signal_state(self, name: str) -> Any:
        """Retrieves a Vega signal's state by name.

        Parameters
        ----------
        name : str
            The name of the signal to retrieve.

        Returns
        -------
        typing.Any
            The current value of the Vega signal.
        """
        return self.server.state[self.ref].get(name, None)

    def update(self, figure: Optional[Chart] = None, **kwargs: Any) -> None:
        super().update(figure=figure, **kwargs)

        if hasattr(self, "_start_update_handlers"):
            self._start_update_handlers()

        self.server.state.flush()



"""Creates the UI for the widget gallery."""

import json
import logging
from asyncio import create_task
from pathlib import Path
from typing import Any, Optional

import blinker
import numpy as np
from altair import Chart, X, Y, selection_interval
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from trame.app import get_server
from trame.decorators import TrameApp
from trame.widgets import client, html
from trame.widgets import vuetify3 as vuetify
from trame_server.core import Server
from trame_server.state import State
from vega_datasets import data

from nova.common.job import ToolOutputs, WorkState
from nova.common.signals import Signal, ToolCommand, get_signal_id
from nova.mvvm.trame_binding import TrameBinding
from nova.trame import ThemedApp
from nova.trame.view.components import DataSelector, FileUpload, InputField, RemoteFileInput
from nova.trame.view.components.execution_buttons import ExecutionButtons
from nova.trame.view.components.ornl import NeutronDataSelector
from nova.trame.view.components.progress_bar import ProgressBar
from nova.trame.view.components.tool_outputs import ToolOutputWindows
from nova.trame.view.components.visualization import Interactive2DPlot, MatplotlibFigure
from nova.trame.view.layouts import GridLayout, HBoxLayout, VBoxLayout
from nova.trame.view.theme.theme import LocalStorageManager

from ..view_models.config import ConfigVM
from ..view_models.data_selector import DataSelectorVM
from ..view_models.file_upload import FileUploadVM
from ..view_models.neutron_data_selector import NeutronDataSelectorVM

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class MplTest:
    """Creates a Matplotlib figure using both Trame options for Matplotlib integration."""

    def __init__(self) -> None:
        self.figure1 = Figure(layout="constrained")
        self.figure2 = Figure(layout="constrained")
        self.ax1 = self.figure1.add_subplot()
        self.ax2 = self.figure2.add_subplot()
        self.fig_type = ""

        self.create_ui()
        self.update()

    def create_ui(self) -> None:
        with GridLayout(columns=2, stretch=True):
            with HBoxLayout(stretch=True):
                self.webagg_view = MatplotlibFigure(self.figure1, webagg=True)
            with HBoxLayout(stretch=True, style="overflow: hidden !important"):
                self.svg_view = MatplotlibFigure(self.figure2)
        vuetify.VBtn("Change MPL Figure", click=self.update)

    def update(self) -> None:
        if self.fig_type == "sin":
            self.fig_type = "cos"
        else:
            self.fig_type = "sin"

        self.plot(self.ax1)
        self.plot(self.ax2)

        self.webagg_view.update(self.figure1)
        self.svg_view.update(self.figure2)

    def plot(self, ax: Axes) -> None:
        ax.clear()

        if self.fig_type == "sin":
            ax.set_title("sin")
            t = np.arange(0.0, 3.0, 0.01)
            s = np.sin(2 * np.pi * t)
        else:
            ax.set_title("cos")
            t = np.arange(0.0, 2.0, 0.01)
            s = np.cos(2 * np.pi * t)

        ax.plot(t, s)


class ComponentTab:
    """Tab for holding the component gallery."""

    def __init__(self, view_model: ConfigVM, local_storage: Optional[LocalStorageManager], **kwargs: Any) -> None:
        self.view_model = view_model
        self.local_storage = local_storage

        self.create_ui(**kwargs)

    def create_ui(self, **kwargs: Any) -> None:
        client.ClientTriggers(mounted=self.read_local_storage)

        with VBoxLayout(classes="pa-16", stretch=True, **kwargs):
            vuetify.VCardTitle("Layouts")
            html.P("GridLayout")
            # [ setup grid ]
            with GridLayout(classes="mb-4", columns=2, halign="center", valign="center"):
                vuetify.VBtn(
                    "{{ item }} - {{ index }}",
                    v_for="(item, index) in ['a', 'b', 'c', 'd']",
                )
            # [ setup grid complete ]
            html.P("HBoxLayout")
            # [ setup hbox ]
            with HBoxLayout(classes="mb-4"):
                vuetify.VBtn("Button 1")
                vuetify.VBtn("Button 2")
                vuetify.VBtn("Button 3")
            # [ setup hbox complete ]
            html.P("VBoxLayout")
            # [ setup vbox ]
            with VBoxLayout(classes="mb-4"):
                vuetify.VBtn("Button 1")
                vuetify.VBtn("Button 2")
                vuetify.VBtn("Button 3")
            # [ setup vbox complete ]

            vuetify.VCardTitle("Advanced Layouts")
            with html.Div(classes="text-left"):
                # [ grid row and column span example ]
                html.P("Grid cells can span multiple rows and/or columns.", classes="text-center")
                with GridLayout(columns=3, height=400, halign="center", valign="center"):
                    # The classes parameter is used to highlight the different cells for demonstration purposes.
                    vuetify.VLabel("Item 1", classes="bg-primary h-100 w-100 justify-center")
                    vuetify.VLabel("Item 2", classes="bg-secondary h-100 w-100 justify-center")
                    vuetify.VLabel("Row Span", classes="bg-primary h-100 w-100 justify-center", row_span=2)
                    vuetify.VLabel("Column Span", classes="bg-error h-100 w-100 justify-center", column_span=2)
                    vuetify.VLabel(
                        "Row and Column Span",
                        classes="bg-secondary h-100 w-100 justify-center",
                        row_span=2,
                        column_span=3,
                    )
                # [ grid row and column span example end ]

                # [ whitespace example ]
                html.P("Grid cells can be left blank.", classes="text-center")
                with GridLayout(columns=3):
                    vuetify.VSheet()
                    InputField(label="Input 1")
                    vuetify.VSheet()
                    InputField(label="Input 2")
                    vuetify.VSheet()
                    InputField(label="Input 3")
                # [ whitespace example end ]

                # [ mixed boxes example 1 ]
                html.P(
                    "HBoxLayout can contain VBoxLayouts.",
                    classes="text-center",
                )
                with HBoxLayout():
                    with VBoxLayout(classes="bg-primary pa-2"):
                        vuetify.VLabel("VBoxLayout 1 - Item 1", classes="mb-4")
                        vuetify.VLabel("VBoxLayout 1 - Item 2")
                    with VBoxLayout(classes="bg-secondary pa-2", halign="center", width=600):
                        vuetify.VLabel("VBoxLayout 2 - Item 1", classes="mb-4")
                        vuetify.VLabel("VBoxLayout 2 - Item 2")
                # [ mixed boxes example 1 end ]

                # [ mixed boxes example 2 ]
                html.P(
                    "VBoxLayout can contain HBoxLayouts.",
                    classes="text-center",
                )
                with VBoxLayout():
                    with HBoxLayout(classes="bg-primary", halign="space-around"):
                        vuetify.VLabel("HBoxLayout 1 - Item 1")
                        vuetify.VLabel("HBoxLayout 1 - Item 2")
                        vuetify.VLabel("HBoxLayout 1 - Item 3")
                    with HBoxLayout(classes="bg-secondary", halign="space-around"):
                        vuetify.VLabel("HBoxLayout 2 - Item 1")
                        vuetify.VLabel("HBoxLayout 2 - Item 2")
                        vuetify.VLabel("HBoxLayout 2 - Item 3")
                # [ mixed boxes example 2 end ]

                # [ mixed boxes example 3 ]
                html.P("GridLayout can also contain HBoxLayouts and VBoxLayouts.", classes="text-center")
                with GridLayout(columns=10):
                    with HBoxLayout(classes="bg-primary", column_span=7, halign="space-around"):
                        vuetify.VLabel("HBoxLayout Item 1")
                        vuetify.VLabel("HBoxLayout Item 2")
                    with VBoxLayout(classes="bg-secondary", column_span=3, halign="center"):
                        vuetify.VLabel("VBoxLayout Item 1")
                        vuetify.VLabel("VBoxLayout Item 2")
                # [ mixed boxes example 3 end ]

            vuetify.VCardTitle("Containment Components")
            with GridLayout(columns=3, halign="center", valign="center"):
                vuetify.VBtn(
                    "Elevated Button",
                )
                vuetify.VBtn(
                    icon="mdi-ab-testing",
                )
                vuetify.VBtn(
                    "Text Button",
                    variant="text",
                )
                with vuetify.VBtnToggle():
                    vuetify.VBtn("Button 1")
                    vuetify.VBtn("Button 2")
                vuetify.VChip("Chip")
                with vuetify.VDialog(max_width=500):
                    with vuetify.Template(v_slot_activator="{ props }"):
                        vuetify.VBtn("Open Dialog", v_bind="props")
                    with vuetify.Template(v_slot_default=True):
                        vuetify.VCard(title="Dialog")
                with html.Div():
                    html.Span("Divider")
                    vuetify.VDivider()
                with vuetify.VExpansionPanels():
                    vuetify.VExpansionPanel(text="Lorem Ipsum", title="Expansion Panel")
                with vuetify.VList():
                    vuetify.VListItem(subtitle="Lorem Ipsum", title="List Item 1")
                    vuetify.VListItem(subtitle="Lorem Ipsum", title="List Item 2")
                    with vuetify.VListItem(subtitle="Lorem Ipsum", title="List Item 3"):
                        # [ Vuetify class example start ]
                        vuetify.VBtn("Button", classes="pr-1 mt-2")
                        # [ Vuetify class example end ]
                with vuetify.VMenu():
                    with vuetify.Template(v_slot_activator="{ props }"):
                        vuetify.VBtn("Open Menu", v_bind="props")
                    with vuetify.VList():
                        vuetify.VListItem("Menu Item")
                with vuetify.VTooltip(text="Tooltip"):
                    with vuetify.Template(v_slot_activator="{ props }"):
                        vuetify.VBtn("Tooltip", v_bind="props")

            vuetify.VCardTitle("Navigation Components")
            with GridLayout(classes="mb-4", columns=2, valign="center"):
                with vuetify.VTabs():
                    vuetify.VTab("Tab 1")
                    vuetify.VTab("Tab 2")
                    vuetify.VTab("Tab 3")

                with vuetify.VTabs(direction="vertical"):
                    vuetify.VTab("Tab 1")
                    vuetify.VTab("Tab 2")

            vuetify.VCardTitle("Data Selection Widgets")
            with GridLayout(classes="mb-1", columns=2, valign="center"):
                InputField(v_model="ds_params.directory")
                InputField(v_model="ds_params.refresh_rate")
            # The flex-shrink statement forces this component to take a pixel height.
            # Better would be to refactor this gallery code fully to work properly as a full-page application,
            # but until that happens please ignore.
            with VBoxLayout(classes="border-md", height=400, style="flex-shrink: 0 !important"):
                DataSelector(
                    v_model="data_selector.selected_files",
                    chips=True,
                    directory=("ds_params.directory", "/"),
                    subdirectory=("ds_params.subdirectory",),
                    refresh_rate=("ds_params.refresh_rate", 15),
                )
            with GridLayout(classes="mb-1", columns=4, valign="center"):
                InputField(v_model="nds_params.facility")
                InputField(v_model="nds_params.instrument")
                InputField(v_model="nds_params.experiment")
                InputField(v_model="nds_params.allow_custom_directories", type="checkbox")
            with VBoxLayout(classes="border-md", height=400, style="flex-shrink: 0 !important"):
                NeutronDataSelector(
                    v_model="nds_filesystem.selected_files",
                    clear_selection_on_experiment_change=False,
                    extensions=[".h5", ".nxs"],
                    facility=("nds_params.facility", "SNS"),
                    instrument=("nds_params.instrument", "TOPAZ"),
                    experiment=("nds_params.experiment", "IPTS-12132"),
                    show_experiment_filters=False,
                    show_selected_files=("false",),
                    allow_custom_directories=("nds_params.allow_custom_directories", True),
                    chips=True,
                )
            with VBoxLayout(classes="border-md", height=400, style="flex-shrink: 0 !important"):
                try:
                    NeutronDataSelector(
                        v_model="nds_oncat.selected_files",
                        data_source="oncat",
                        projection=["indexed.run_number", "metadata.entry.title"],
                        chips=True,
                    )
                except EnvironmentError:
                    # Unit tests will fail without this in the CI since ONCat vars won't be defined.
                    html.Span(
                        "You must define ONCAT_CLIENT_ID and ONCAT_CLIENT_SECRET in your environment in order "
                        "to test the ONCat backend for NeutronDataSelector."
                    )

            vuetify.VCardTitle("Form Inputs & Controls")
            with GridLayout(columns=3, valign="center"):
                FileUpload(
                    v_model="file_upload.file",
                    base_paths=("/HFIR", "/SNS"),
                    extensions="file_upload.extensions",
                    label="{{ file_upload.label }}",
                    # show_server_files=False,
                )
                with html.Div():
                    InputField(
                        v_model="config.autoscroll",
                        auto_grow=True,
                        classes="mb-2",
                        label="Autoscroll Text Area",
                        max_rows=5,
                        rows=1,
                        type="autoscroll",
                    )
                    vuetify.VBtn("Add Line to Autoscroller", click=self.append_to_autoscroll)
                InputField(type="checkbox", label="Checkbox")
                InputField(type="radio", label="Radio Buttons", inline=True, items=("config.radio_items",))
                InputField(type="file", label="File Upload")
                InputField(
                    type="select",
                    items="['Option 1', 'Option 2']",
                    label="Select",
                )
                InputField(type="slider", label="Slider")
                InputField(type="switch", label="Switch")
                # [ InputField kwargs example start ]
                InputField(type="textarea", auto_grow=True, label="Text Area")
                # [ InputField kwargs example end ]
                InputField(ref="pydantic-field", id="pydantic-field", v_model="config.value")
                InputField(v_model=("config.debounce_rate",))
                InputField(v_model=("config.debounce",), debounce=("config.debounce_rate",))
                InputField(v_model=("config.throttle",), throttle=1000)
                InputField(type="number", label="Number Input")
                RemoteFileInput(
                    v_model="config.selected_file",
                    base_paths=["/run"],
                    extensions=[".pid", ".lock"],
                    input_props={"label": "File Selector"},
                )
                RemoteFileInput(
                    v_model="config.selected_folder",
                    allow_files=False,
                    allow_folders=True,
                    base_paths=["/usr"],
                    input_props={"label": "Folder Selector"},
                )

            vuetify.VCardTitle("Validation")
            with vuetify.VForm():
                with GridLayout(classes="mb-4", columns=3, width=600, valign="center"):
                    InputField(label="Required Field", required=True)
                    InputField(label="Optional Field")
                    InputField(
                        label="Text Only Optional Field",
                        rules=("[(value) => /[0-9]/.test(value) ? 'Field must not contain numbers' : true]",),
                    )
                    InputField(
                        ref="gallery_select",
                        v_model="config.select1",
                        type="select",
                        items="['Option 1', 'Option 2']",
                        label="Required Select",
                        multiple=True,
                    )
                    InputField(
                        v_model="config.select2",
                        type="select",
                        items="['Option 1', 'Option 2']",
                        label="Cross-validated Select",
                        multiple=True,
                    )

            vuetify.VCardTitle("Feedback Components")
            with GridLayout(columns=3, halign="center", valign="center"):
                vuetify.VAlert("Alert")
                with vuetify.VBadge():
                    vuetify.VIcon("mdi-ab-testing")
                vuetify.VProgressCircular(indeterminate=True)
                vuetify.VProgressLinear(indeterminate=True)
                with vuetify.VSnackbar(
                    "Snackbar",
                    v_model="snackbar",
                    timeout=-1,
                ):
                    with vuetify.Template(v_slot_actions=True):
                        vuetify.VBtn("Close", click="snackbar = false")

            vuetify.VCardTitle("Visualization Components")
            self.plot = Interactive2DPlot(
                figure=self.create_test_2d_plot(),
                id="interactive-plot",
                __events=["pointerup"],
                pointerup=self.read_plot_signal,
            )
            html.P(f"Selected interval: {{{{ {self.plot.ref}['interval'] }}}}")

            vuetify.VCardTitle("Local Storage")
            with html.Div():
                InputField(
                    v_model="local_storage.value",
                    classes="mb-2 mt-0 mx-auto",
                    id="local-storage-input",
                    label="Local Storage Test",
                    width=400,
                )
                vuetify.VBtn(
                    "Save to LocalStorage", classes="mr-2", id="local-storage-set", click=self.set_local_storage
                )
                vuetify.VBtn("Clear LocalStorage", id="local-storage-remove", click=self.remove_local_storage)

            ProgressBar("test")
            ToolOutputWindows("test")
            ExecutionButtons("test")
            ExecutionButtons("test_bindings", stop_btn=("true",), download_btn=False)

    def create_test_2d_plot(self) -> Chart:
        brush = selection_interval(encodings=["x"], name="interval", zoom=False)

        return (
            Chart(data.cars(), title="Interactive 2D Plot")
            .mark_circle()
            .encode(
                X("Horsepower", type="quantitative"),
                Y("Miles_per_Gallon", type="quantitative"),
                color="Origin:N",
            )
            .properties(height=400, width=1000)
            .add_params(brush)
        )

    def read_plot_signal(self) -> None:
        logger.info(f"Interval signal state: {self.plot.get_signal_state('interval')}")

    async def _read_local_storage(self) -> None:
        if self.local_storage:
            self.view_model.set_local_storage(await self.local_storage.get("local_storage_test"))

    def read_local_storage(self) -> None:
        create_task(self._read_local_storage())

    def remove_local_storage(self) -> None:
        if self.local_storage:
            self.local_storage.remove("local_storage_test")

    def set_local_storage(self) -> None:
        if self.local_storage:
            self.local_storage.set("local_storage_test", self.view_model.get_local_storage())

    def append_to_autoscroll(self) -> None:
        self.view_model.append_to_autoscroll("Line added by button\n")


class FullScreenTab:
    """Tab showing MatplotlibFigure in our full-screen layout."""

    def __init__(self, **kwargs: Any) -> None:
        self.create_ui(**kwargs)

    def create_ui(self, **kwargs: Any) -> None:
        with VBoxLayout(stretch=True, **kwargs):
            MplTest()


@TrameApp()
class App(ThemedApp):
    """Root Trame class for defining the UI."""

    def __init__(self, server: Server = None) -> None:
        """Constructor for the App class."""
        try:
            with open(Path(__file__).parent / "vuetify_config.json", "r") as _file:
                vuetify_config = json.load(_file)
                logger.warning(
                    "WARN: Gallery loaded a local Vuetify config. This is only provided as an example and should not "
                    "be used in production."
                )
        except (FileNotFoundError, ValueError):
            vuetify_config = {}
        super().__init__(server=server, vuetify_config_overrides=vuetify_config)

        self.server = get_server(server, client_type="vue3")
        command_signal = blinker.signal(get_signal_id("test", Signal.TOOL_COMMAND))
        command_signal.connect(self._set_state)
        self.create_state()
        self.create_ui()

    @property
    def state(self) -> State:
        return self.server.state

    async def _set_state(self, _sender: Any, command: str) -> None:
        progress_signal = blinker.signal(get_signal_id("test", Signal.PROGRESS))
        outputs_signal = blinker.signal(get_signal_id("test", Signal.OUTPUTS))

        if command == ToolCommand.START:
            await progress_signal.send_async(
                "test_sender",
                state=WorkState.RUNNING,
                details={
                    "message": "",
                    "original_dict": {
                        "test": "longstring longstring longstring longstring longstring",
                        "test2": "value2",
                    },
                },
            )
            await outputs_signal.send_async(
                "test_sender", outputs=ToolOutputs(stdout="test_output", stderr="test_error")
            )
        else:
            await progress_signal.send_async("test_sender", state=WorkState.FINISHED, details={})

    def create_state(self) -> None:
        binding = TrameBinding(self.state)

        self.data_selector_vm = DataSelectorVM(binding)
        self.data_selector_vm.model_bind.connect("data_selector")
        self.data_selector_vm.parameter_bind.connect("ds_params")

        self.neutron_data_selector_vm = NeutronDataSelectorVM(binding)
        self.neutron_data_selector_vm.analysis_bind.connect("nds_filesystem")
        self.neutron_data_selector_vm.oncat_bind.connect("nds_oncat")
        self.neutron_data_selector_vm.parameter_bind.connect("nds_params")

        self.file_upload_vm = FileUploadVM(binding)
        self.file_upload_vm.model_bind.connect("file_upload")

        self.config_vm = ConfigVM(binding)
        self.config_vm.config_bind.connect("config")
        self.config_vm.local_storage_bind.connect("local_storage")

    def create_ui(self) -> None:
        self.state.trame__title = "Widget Gallery"
        with super().create_ui() as layout:
            layout.toolbar_title.set_text("Widget Gallery")

            # [slot child example]
            with layout.actions:
                vuetify.VBtn("Text Button")
            # [slot child example complete]

            with layout.pre_content:
                with client.DeepReactive("config"):
                    with vuetify.VTabs(v_model="config.active_tab", classes="pl-6"):
                        vuetify.VTab("Components", value=0)
                        vuetify.VTab("Full-screen Layout", value=1)

            with layout.content:
                ComponentTab(self.config_vm, self.local_storage, v_if="config.active_tab == 0")
                FullScreenTab(v_if="config.active_tab == 1")

            with layout.post_content:
                html.Div("Sticky Bottom Content", classes="text-center w-100")

            return layout



"""Components used to control the lifecycle of a Themed Application."""

import logging
from asyncio import sleep
from typing import Any

from trame.app import get_server
from trame.widgets import client
from trame.widgets import vuetify3 as vuetify

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class ExitButton:
    """Exit button for Trame Applications."""

    def __init__(self, exit_callback: Any = None, job_status_callback: Any = None) -> None:
        self.server = get_server(None, client_type="vue3")
        self.server.state.nova_kill_jobs_on_exit = False
        self.server.state.nova_show_exit_dialog = False
        self.server.state.nova_show_stop_jobs_on_exit_checkbox = False
        self.server.state.nova_running_jobs = []
        self.server.state.nova_show_exit_progress = False
        # Note that window.close() will fail in many situations due to security concerns: https://developer.mozilla.org/en-US/docs/Web/API/Window/close
        # This is simply a best effort to close the tab. My hope is that it will generally work when people are running
        # applications through our dashboard since tabs are opened via target="_blank" links.
        self.close_browser = client.JSEval(exec="window.close();").exec
        self.exit_application_callback = exit_callback
        self.job_status_callback = job_status_callback
        self.create_ui()

    def create_ui(self) -> None:
        with vuetify.VBtn(
            "Exit",
            prepend_icon="mdi-close-box",
            classes="exit-button mr-4 bg-secondary",
            id="shutdown_app_theme_button",
            color="white",
            size="default",
            click=self.open_exit_dialog,
        ):
            with vuetify.VDialog(v_model="nova_show_exit_dialog", persistent="true"):
                with vuetify.VCard(classes="pa-4 ma-auto"):
                    vuetify.VCardTitle("Exit Application")
                    with vuetify.VCardText(
                        "Are you sure you want to exit this application?",
                        variant="outlined",
                    ):
                        vuetify.VCheckbox(
                            v_model="nova_kill_jobs_on_exit",
                            label="Stop All Jobs On Exit.",
                            v_if="nova_running_jobs.length > 0",
                        )
                        with vuetify.VList():
                            vuetify.VListSubheader("Running Jobs:", v_if="nova_running_jobs.length > 0")
                            vuetify.VListItem("{{ item }}", v_for="(item, index) in nova_running_jobs")
                    with vuetify.VCardActions(v_if="!nova_show_exit_progress"):
                        vuetify.VBtn(
                            "Exit App",
                            click=self.exit_application,
                            color="error",
                        )
                        vuetify.VBtn(
                            "Stay In App",
                            click=self.close_exit_dialog,
                        )
                    with vuetify.VCardActions(v_else=True):
                        vuetify.VCardText(
                            "Exiting Application...",
                            variant="outlined",
                        )
                        vuetify.VProgressCircular(indeterminate=True)

    async def exit_application(self) -> None:
        self.close_browser()

        # sleep gives time for the Trame server to communicate the close request to the browser.
        await sleep(0.1)
        await self.exit_application_callback()

    async def open_exit_dialog(self) -> None:
        self.server.state.nova_show_exit_dialog = True
        await self.job_status_callback()

    async def close_exit_dialog(self) -> None:
        self.server.state.nova_show_exit_dialog = False



"""Internal utilities for nova-trame."""

from typing import Any, Tuple, Union

from trame_server.core import State

from nova.mvvm._internal.utils import rgetdictvalue, rsetdictvalue


# Trame state handlers don't work on nested properties. When writing Trame state handlers (e.g. flushState, dirty, or
# change), we instead use the name of the top-level property. For example, "config.parameter_group_a.option_x" becomes
# "config".
def get_state_name(name: str) -> str:
    return name.split(".")[0]


# Reads a state parameter from Trame. For internal use only, if you're using this in your application you're violating
# the MVVM framework. :)
def get_state_param(state: State, value: Union[Any, Tuple]) -> Any:
    if isinstance(value, tuple):
        return rgetdictvalue(state, value[0])

    return value


# Writes a state parameter to Trame. For internal use only, if you're using this in your application you're violating
# the MVVM framework. :)
def set_state_param(state: State, value: Union[Any, Tuple], new_value: Any = None) -> Any:
    with state:
        if isinstance(value, tuple):
            if new_value is not None:
                rsetdictvalue(state, value[0], new_value)
            elif len(value) > 1:
                rsetdictvalue(state, value[0], value[1])
            state.dirty(get_state_name(value[0]))

    return get_state_param(state, value)



"""Model state for RemoteFileInput."""

import os
from functools import cmp_to_key
from locale import strcoll
from typing import Any, List, Union

from pydantic import BaseModel, Field


class RemoteFileInputState(BaseModel):
    """Pydantic model for RemoteFileInput state."""

    allow_files: bool = Field(default=False)
    allow_folders: bool = Field(default=False)
    base_paths: List[str] = Field(default=[])
    extensions: List[str] = Field(default=[])


class RemoteFileInputModel:
    """Manages interactions between RemoteFileInput and the file system."""

    def __init__(self) -> None:
        """Creates a new RemoteFileInputModel."""
        self.state = RemoteFileInputState()

    def set_binding_parameters(self, **kwargs: Any) -> None:
        if "allow_files" in kwargs:
            self.state.allow_files = kwargs["allow_files"]
        if "allow_folders" in kwargs:
            self.state.allow_folders = kwargs["allow_folders"]
        if "base_paths" in kwargs:
            self.state.base_paths = kwargs["base_paths"]
        if "extensions" in kwargs:
            self.state.extensions = kwargs["extensions"]

    def get_base_paths(self) -> list[dict[str, Any]]:
        return [{"path": base_path, "directory": True} for base_path in self.state.base_paths]

    def scan_current_path(
        self, current_path: str, showing_all_files: bool, filter: str
    ) -> tuple[list[dict[str, Any]], bool]:
        failed = False
        files = []
        filter = filter.split("/")[-1]

        try:
            if current_path and (not self.valid_subpath(current_path) or not os.path.exists(current_path)):
                raise FileNotFoundError

            files = [{"path": "..", "directory": True}]

            if os.path.isdir(current_path):
                scan_path = current_path
            else:
                scan_path = os.path.dirname(current_path)

            for entry in os.scandir(scan_path):
                if self.valid_entry(entry, showing_all_files) and (not filter or entry.name.startswith(filter)):
                    files.append({"path": entry.name, "directory": entry.is_dir()})
        except OSError:
            if not current_path or current_path == "/":
                files = self.get_base_paths()
                failed = True

        def _sort_files(a: dict[str, Any], b: dict[str, Any]) -> int:
            if a["directory"] and not b["directory"]:
                return -1
            if b["directory"] and not a["directory"]:
                return 1

            path_a = a["path"].lower()
            path_b = b["path"].lower()

            return strcoll(path_a, path_b)

        sorted_files = sorted(files, key=cmp_to_key(_sort_files))

        return (sorted_files, failed)

    def select_file(self, file: Union[dict[str, str], str], old_path: str, showing_base_paths: bool) -> str:
        if file == "":
            return ""

        if isinstance(file, dict):
            file = file["path"]

        if not os.path.isdir(old_path):
            # If the previous selection is a file, then we need to append to its parent directory
            old_path = os.path.dirname(old_path)

        if not showing_base_paths and file != "..":
            return os.path.join(old_path, file)
        elif not showing_base_paths:
            if old_path in self.state.base_paths:
                return ""
            else:
                return os.path.dirname(old_path)
        else:
            return file

    def valid_entry(self, entry: os.DirEntry, showing_all_files: bool) -> bool:
        if entry.is_dir():
            return True

        if not self.state.allow_files:
            return False

        return (
            showing_all_files
            or not self.state.extensions
            or any(entry.name.endswith(ext) for ext in self.state.extensions)
        )

    def valid_selection(self, selection: str) -> bool:
        if self.valid_subpath(selection):
            if os.path.isdir(selection) and self.state.allow_folders:
                return True

            if os.path.isfile(selection) and self.state.allow_files:
                return True

        return False

    def valid_subpath(self, subpath: str) -> bool:
        if subpath == "":
            return False

        for base_path in self.state.base_paths:
            if subpath.startswith(base_path):
                return True

        return False



"""Pydantic model for testing validation."""

from typing import Dict, List, Union

from pydantic import BaseModel, Field, field_validator


class Config(BaseModel):
    """Pydantic model for testing validation."""

    autoscroll: str = Field(default="", title="Autoscroller")
    nested: Dict[str, str] = Field(default={"selected_file": ""})
    select1: List[str] = Field(default=[], title="Select")
    select2: List[str] = Field(default=[], title="Select")
    selected_file: str = Field(default="", title="Selected File")
    selected_folder: str = Field(default="", title="Selected Folder")
    snackbar: bool = Field(default=True)
    active_tab: int = Field(default=0)
    debounce_rate: int = Field(default=1000, title="Debounce Rate")
    debounce: str = Field(
        default="",
        description="This field is debounced and will not update its state until you've stopped typing for 1 second.",
        title="Debounced Field",
    )
    radio_items: List[Dict[str, Union[str, int]]] = Field(
        default=[{"title": "Item 1", "value": 1}, {"title": "Item 2", "value": 2}]
    )
    throttle: str = Field(
        default="",
        description="This field is throttled and will only update its state every 1 second.",
        title="Throttled Field",
    )
    value: int = Field(default=0, description="This field is validated via Pydantic.", title="Pydantic Field")

    @field_validator("debounce", mode="after")
    @classmethod
    def on_debounce(cls, text: str) -> str:
        if text:
            print(f"received debounced update: {text}")

        return text

    @field_validator("throttle", mode="after")
    @classmethod
    def on_throttle(cls, text: str) -> str:
        if text:
            print(f"received throttled update: {text}")

        return text


class LocalStorageState(BaseModel):
    """Separate Pydantic model to hold local storage state which can leak in unit tests."""

    value: str = Field(default="", title="Local Storage")



"""Unit tests for LocalStorageManager."""

from time import sleep

from selenium.webdriver import Firefox
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions
from selenium.webdriver.support.ui import WebDriverWait


def test_local_storage(driver: Firefox) -> None:
    assert driver.execute_script("window.localStorage.getItem('local_storage_test');") is None

    text_field = WebDriverWait(driver, 10).until(
        expected_conditions.element_to_be_clickable((By.CSS_SELECTOR, "#local-storage-input"))
    )
    text_field.send_keys("1")
    sleep(1)
    driver.execute_script("window.document.getElementById('local-storage-set').click();")
    WebDriverWait(driver, 10).until(
        lambda *args: driver.execute_script("return window.localStorage.getItem('local_storage_test');") == "1"
    )

    driver.execute_script("window.document.getElementById('local-storage-remove').click();")
    assert driver.execute_script("window.localStorage.getItem('local_storage_test');") is None



"""Test the UI via Selenium."""

from selenium.webdriver import Firefox
from selenium.webdriver.common.by import By


def test_interface(driver: Firefox) -> None:
    title = driver.find_element(By.CLASS_NAME, "v-toolbar-title__placeholder").get_attribute("innerHTML")

    assert title
    assert "Widget Gallery" in title



"""Unit tests for MatplotlibFigure."""

from matplotlib.figure import Figure

from nova.trame.view.components.visualization import MatplotlibFigure


def test_matplotlib() -> None:
    mpl_figure = Figure()

    svg_test = MatplotlibFigure()
    assert svg_test._webagg is False

    svg_test.update(None)
    svg_test.update(mpl_figure)

    webagg_test = MatplotlibFigure(webagg=True)
    assert webagg_test._webagg is True

    webagg_test.update(None)
    webagg_test.update(mpl_figure)



"""Unit tests for InputField."""

from enum import Enum
from time import sleep
from typing import cast

from pydantic import BaseModel, Field
from selenium.webdriver import Firefox
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions
from selenium.webdriver.support.ui import WebDriverWait
from trame.app import get_server
from trame.widgets import vuetify3 as vuetify
from trame_server import Server

from nova.mvvm.trame_binding import TrameBinding
from nova.trame import ThemedApp
from nova.trame.view.components import InputField


def test_input_field() -> None:
    input_field = InputField(type="autocomplete")
    assert isinstance(input_field, vuetify.VAutocomplete)

    input_field = InputField(type="checkbox")
    assert isinstance(input_field, vuetify.VCheckbox)

    input_field = InputField(type="combobox")
    assert isinstance(input_field, vuetify.VCombobox)

    input_field = InputField(type="file")
    assert isinstance(input_field, vuetify.VFileInput)

    input_field = InputField(type="input")
    assert isinstance(input_field, vuetify.VInput)

    input_field = InputField(type="number")
    assert isinstance(input_field, vuetify.VNumberInput)

    input_field = InputField(type="otp")
    assert isinstance(input_field, vuetify.VOtpInput)

    input_field = InputField(type="radio")
    assert isinstance(input_field, vuetify.VRadioGroup)

    input_field = InputField(type="range-slider")
    assert isinstance(input_field, vuetify.VRangeSlider)

    input_field = InputField(type="select")
    assert isinstance(input_field, vuetify.VSelect)

    input_field = InputField(type="slider")
    assert isinstance(input_field, vuetify.VSlider)

    input_field = InputField(type="switch")
    assert isinstance(input_field, vuetify.VSwitch)

    input_field = InputField(type="textarea")
    assert isinstance(input_field, vuetify.VTextarea)

    input_field = InputField(type="password")
    assert isinstance(input_field, vuetify.VTextField)


def test_pydantic() -> None:
    class User(BaseModel):
        username: str = Field(
            default="test_name", min_length=1, title="User Name", description="hint", examples=["user"]
        )

    obj = User()

    class MyTrameApp(ThemedApp):
        def __init__(self, server: Server = None) -> None:
            server = get_server(None, client_type="vue3")
            super().__init__(server=server)
            binding = TrameBinding(server.state).new_bind(obj)
            binding.connect("obj")
            self.create_ui()

        def create_ui(self) -> None:
            with super().create_ui() as layout:
                with layout.content:
                    input_field = cast(vuetify.VTextField, InputField(v_model="obj.username"))
                    assert input_field.label == "User Name"
                    assert input_field.placeholder == "user"
                    assert input_field.title == "hint"

    MyTrameApp()


def test_pydantic_validation(driver: Firefox) -> None:
    sleep(1)
    input_field = WebDriverWait(driver, 10).until(
        expected_conditions.element_to_be_clickable((By.CSS_SELECTOR, "#pydantic-field"))
    )
    input_field.send_keys("test")
    sleep(1)
    error_message = driver.execute_script("""
        const messages_content = document.getElementById("pydantic-field-messages");

        return messages_content.textContent;
    """)

    assert error_message.startswith("Input should be a valid integer")


def test_items_attributes() -> None:
    # items autopopulation start
    class DropdownOptions(str, Enum):
        item_a = "item_a"
        item_b = "item_b"
        item_c = "item_c"

    class Dropdown(BaseModel):
        enum_field: DropdownOptions = Field(default=DropdownOptions.item_a)
        str_field: str = Field(default="test")

    dropdown = Dropdown()
    # items autopopulation end

    class MyTrameApp(ThemedApp):
        def __init__(self, server: Server = None) -> None:
            server = get_server(None, client_type="vue3")
            super().__init__(server=server)
            binding = TrameBinding(server.state).new_bind(dropdown)
            binding.connect("dropdown")
            self.create_ui()

        def create_ui(self) -> None:
            with super().create_ui() as layout:
                with layout.content:
                    autopopulated_items = cast(
                        vuetify.VSelect, InputField(v_model="dropdown.enum_field", type="select")
                    )
                    missing_select_type = cast(vuetify.VSelect, InputField(v_model="dropdown.enum_field"))
                    missing_enum_model = cast(vuetify.VSelect, InputField(v_model="dropdown.str_field", type="select"))

                    assert autopopulated_items.items == ("['item_a', 'item_b', 'item_c']",)
                    assert missing_select_type.items is None
                    assert missing_enum_model.items is None

    MyTrameApp()


def test_throttle_attributes() -> None:
    class MyTrameApp(ThemedApp):
        def __init__(self, server: Server = None) -> None:
            server = get_server(None, client_type="vue3")
            super().__init__(server=server)

            self.create_ui()

        def create_ui(self) -> None:
            with super().create_ui() as layout:
                with layout.content:
                    debounced_input = InputField(v_model="debounce_test", debounce=1000)
                    throttled_input = InputField(v_model="throttle_test", throttle=2000)
                    noop_input = InputField(
                        debounce=1000
                    )  # There's no state to update here, so update_modelValue should still be None.

                    assert hasattr(debounced_input, "update_modelValue")
                    assert debounced_input.update_modelValue.startswith("window.delay_manager.debounce")
                    assert hasattr(throttled_input, "update_modelValue")
                    assert throttled_input.update_modelValue.startswith("window.delay_manager.throttle")
                    assert hasattr(noop_input, "update_modelValue")
                    assert noop_input.update_modelValue is None

                    try:
                        InputField(v_model="invalid_input", debounce=1000, throttle=1000)
                        raise AssertionError("Debounce and throttle used together should raise a ValueError")
                    except ValueError:
                        pass

    MyTrameApp()


def test_invalid_rules() -> None:
    try:
        InputField(rules=42)
        raise AssertionError("Expected ValueError from invalid rules list")
    except ValueError:
        pass


def test_help() -> None:
    input_field = cast(
        vuetify.VTextField, InputField(help={"hint": "This is a hint.", "placeholder": "This is a placeholder."})
    )
    assert input_field.placeholder == "This is a placeholder."
    assert input_field.title == "This is a hint."


def test_required() -> None:
    input_field = cast(vuetify.VTextField, InputField(required=True))
    assert input_field.label == "*"


def test_change_handlers() -> None:
    input_field = cast(vuetify.VTextField, InputField(change=lambda: print("hi")))
    assert input_field.change.startswith(f"trigger('{input_field.ref}__trigger', [], {{}});")

    input_field = cast(vuetify.VTextField, InputField(change="console.log('hi');"))
    assert input_field.change.startswith("console.log('hi');")



"""Unit tests for NeutronDataSelector."""

from typing import List
from warnings import catch_warnings

from pydantic import BaseModel, Field
from trame.app import get_server
from trame_server.core import Server

from nova.mvvm.trame_binding import TrameBinding
from nova.trame.view.components.ornl import NeutronDataSelector
from nova.trame.view.theme import ThemedApp


def test_data_selector() -> None:
    class MyTrameApp(ThemedApp):
        def __init__(self, server: Server = None) -> None:
            server = get_server(None, client_type="vue3")
            super().__init__(server=server)
            self.create_ui()

        def create_ui(self) -> None:
            with super().create_ui() as layout:
                with layout.content:
                    input = NeutronDataSelector(v_model="test", extensions=[".tiff"])
                    assert input.v_model == "test"
                    assert input._extensions == [".tiff"]
                    assert input._model.state.facility == ""
                    assert input._model.state.instrument == ""
                    assert input._model.state.experiment == ""

                    input.update_facility("HFIR")
                    input.update_instrument("BIO-SANS")
                    input.update_experiment("IPTS-24666")

                    with catch_warnings(record=True) as captured_warnings:
                        input.update_facility(facility="NSS")
                        assert str(captured_warnings[0].message).startswith("Facility 'NSS' could not be found.")

                    with catch_warnings(record=True) as captured_warnings:
                        NeutronDataSelector(v_model="test", facility="HIFR")
                        assert str(captured_warnings[0].message).startswith("Facility 'HIFR' could not be found.")

                    with catch_warnings(record=True) as captured_warnings:
                        NeutronDataSelector(v_model="test", facility="SNS", instrument="BL1B")
                        assert str(captured_warnings[0].message).startswith(
                            "Instrument 'BL1B' could not be found in 'SNS'."
                        )

                    try:
                        input.set_state()
                        raise AssertionError("set_state should fail to run")
                    except TypeError:
                        pass

    MyTrameApp()


def test_parameter_bindings() -> None:
    class TestModel(BaseModel):
        v_model: List[str] = Field(default=[])
        facility: str = Field(default="")
        instrument: str = Field(default="")
        experiment: str = Field(default="")
        allow_custom_directories: bool = Field(default=False)

    class MyTrameApp(ThemedApp):
        def __init__(self, server: Server = None) -> None:
            self.server = get_server(None, client_type="vue3")
            super().__init__(server=self.server)

            self.create_binding()
            self.create_ui()

        def create_binding(self) -> None:
            self.test_obj = TestModel()

            binding = TrameBinding(self.server.state)
            self.test_binding = binding.new_bind(self.test_obj)
            self.test_binding.connect("test_nds")

        def create_ui(self) -> None:
            with super().create_ui() as layout:
                with layout.content:
                    NeutronDataSelector(
                        v_model=("test_nds.v_model", ["test.txt"]),
                        facility=("test_nds.facility", "SNS"),
                        instrument=("test_nds.instrument", "TOPAZ"),
                        experiment=("test_nds.experiment",),
                        allow_custom_directories=("test_nds.allow_custom_directories", True),
                    )

    MyTrameApp()



"""Unit tests for GridLayout, HBoxLayout, and VBoxLayout."""

from trame.widgets import html
from trame.widgets import vuetify3 as vuetify

from nova.trame.view.layouts import GridLayout, HBoxLayout, VBoxLayout


def test_grid() -> None:
    with GridLayout(columns=3, stretch=True):
        html.Div("Test 1")
        html.Div("Test 2", row_span=2, column_span=2)


def test_complex_layout() -> None:
    # [ setup complex layout example ]
    class LMRLayout:
        def __init__(self) -> None:
            with GridLayout(rows=1, columns=10, halign="center", valign="center"):
                self.left = html.Div(column_span=2)  # 20% width
                self.middle = html.Div(column_span=5)  # 50% width
                self.right = html.Div(column_span=3)  # 30% width

    my_layout = LMRLayout()
    with my_layout.left:
        vuetify.VBtn("Left Button")
    # [ setup complex layout example complete ]


def test_hbox() -> None:
    with HBoxLayout(stretch=True):
        html.Div("Test 1")
        html.Div("Test 2")


def test_vbox() -> None:
    with VBoxLayout(stretch=True):
        html.Div("Test 1")
        html.Div("Test 2")


def test_style_attributes() -> None:
    # Test dict-typed style
    with GridLayout(style={"height": "40px"}):
        html.Div("Test", style={"height": "100%"})
    with HBoxLayout(style={"height": "40px"}):
        html.Div("Test", style={"height": "100%"})
    with VBoxLayout(style={"height": "40px"}):
        html.Div("Test", style={"height": "100%"})

    # Test string-typed style
    with GridLayout(style="height: 40px"):
        html.Div("Test", style="height: 40px")
    with HBoxLayout(style="height: 40px"):
        html.Div("Test", style="height: 40px")
    with VBoxLayout(style="height: 40px"):
        html.Div("Test", style="height: 40px")



"""Implementation of LocalStorageManager."""

from asyncio import sleep
from typing import Any, Optional

from trame.widgets import client
from trame_server.core import Controller


class LocalStorageManager:
    """Allows manipulation of window.localStorage from your Python code.

    LocalStorageManager requires a Trame layout to exist in order to work properly. Because of this, it's strongly
    recommended that you don't use this class directly. Instead, ThemedApp automatically creates an instance of this
    class and stores it in ThemedApp.local_storage, through which it can safely be used.
    """

    def __init__(self, ctrl: Controller) -> None:
        """Constructor for the LocalStorageManager class.

        Parameters
        ----------
        ctrl : `trame_server.core.Controller <https://trame.readthedocs.io/en/latest/core.controller.html#trame_server.controller.Controller>`_
            The Trame controller.

        Returns
        -------
        None
        """
        self.js_get = client.JSEval(
            exec=(
                "window.trame.trigger("
                "  'nova__local_storage_trigger', "
                "  [$event.key, window.localStorage.getItem($event.key)]"
                ");"
            )
        ).exec
        self.js_remove = client.JSEval(exec="window.localStorage.removeItem($event.key);").exec
        self.js_set = client.JSEval(exec="window.localStorage.setItem($event.key, $event.value);").exec

        self._ready: dict[str, bool] = {}
        self._values: dict[str, str] = {}

        @ctrl.trigger("nova__local_storage_trigger")
        def _(key: str, value: str) -> None:
            self._ready[key] = True
            self._values[key] = value

    async def get(self, key: str) -> Optional[str]:
        """Gets the value of a key from window.localStorage.

        You cannot call this from the main Trame coroutine because this waits on a response from the browser that must
        be processed by the main coroutine. Instead, you should call this from another thread or coroutine, typically
        with :code:`asyncio.create_task`.

        Parameters
        ----------
        key : str
            The key to get the value of.

        Returns
        -------
        Optional[str]
            The value of the key from window.localStorage.
        """
        self._ready[key] = False
        self.js_get({"key": key})

        while not self._ready[key]:
            await sleep(0.1)

        return self._values[key]

    def remove(self, key: str) -> None:
        """Removes a key from window.localStorage.

        Parameters
        ----------
        key : str
            The key to remove.

        Returns
        -------
        None
        """
        self.js_remove({"key": key})

    def set(self, key: str, value: Any) -> None:
        """Sets the value of a key in window.localStorage.

        Parameters
        ----------
        key : str
            The key to set the value of.
        value : typing.Any
            The value to set. This value will be coerced to a string before being stored.

        Returns
        -------
        None
        """
        self.js_set({"key": key, "value": value})



"""Analysis cluster filesystem backend for NeutronDataSelector."""

import os
from pathlib import Path
from typing import Any, Dict, List, Optional
from warnings import warn

from natsort import natsorted
from pydantic import Field, model_validator
from typing_extensions import Self

from .neutron_data_selector import NeutronDataSelectorModel, NeutronDataSelectorState

CUSTOM_DIRECTORIES_LABEL = "Custom Directory"

INSTRUMENT_DIRS = {
    "HFIR": {
        "CG-1A": "CG1A",
        "DEV BEAM": "CG1B",
        "MARS": "CG1D",
        "GP-SANS": "CG2",
        "BIO-SANS": "CG3",
        "CNPDB": "CG4B",
        "CTAX": "CG4C",
        "IMAGINE": "CG4D",
        "PTAX": "HB1",
        "VERITAS": "HB1A",
        "POWDER": "HB2A",
        "HIDRA": "HB2B",
        "WAND²": "HB2C",
        "TAX": "HB3",
        "DEMAND": "HB3A",
        "NOWG": "NOWG",
        "NOWV": "NOWV",
    },
    "SNS": {
        "ARCS": "ARCS",
        "BL-0": "BL0",
        "BASIS": "BSS",
        "CNCS": "CNCS",
        "CORELLI": "CORELLI",
        "EQ-SANS": "EQSANS",
        "HYSPEC": "HYS",
        "MANDI": "MANDI",
        "NOMAD": "NOM",
        "NOWB": "NOWB",
        "NOWD": "NOWD",
        "NSE": "NSE",
        "POWGEN": "PG3",
        "LIQREF": "REF_L",
        "MAGREF": "REF_M",
        "SEQUOIA": "SEQ",
        "SNAP": "SNAP",
        "TOPAZ": "TOPAZ",
        "USANS": "USANS",
        "VENUS": "VENUS",
        "VISION": "VIS",
        "VULCAN": "VULCAN",
    },
}

INSTRUMENT_IDS = {
    "HFIR": {
        "CG-1A": "CG-1A",
        "DEV BEAM": "CG-1B",
        "MARS": "CG-1D",
        "GP-SANS": "CG-2",
        "BIO-SANS": "CG-3",
        "CNPDB": "CG-4B",
        "CTAX": "CG-4C",
        "IMAGINE": "CG-4D",
        "PTAX": "HB-1",
        "VERITAS": "HB-1A",
        "POWDER": "HB-2A",
        "HIDRA": "HB-2B",
        "WAND²": "HB-2C",
        "TAX": "HB-3",
        "DEMAND": "HB-3A",
        "NOWG": "NOW-G",
        "NOWV": "NOW-V",
    },
    "SNS": {
        "ARCS": "BL-18",
        "BL-0": "BL-0",
        "BASIS": "BL-2",
        "CNCS": "BL-5",
        "CORELLI": "BL-9",
        "EQ-SANS": "BL-6",
        "HYSPEC": "BL-14B",
        "MANDI": "BL-11B",
        "NOMAD": "BL-1B",
        "NOWB": "NOW-B",
        "NOWD": "NOW-D",
        "NSE": "BL-15",
        "POWGEN": "BL-11A",
        "LIQREF": "BL-4B",
        "MAGREF": "BL-4A",
        "SEQUOIA": "BL-17",
        "SNAP": "BL-3",
        "TOPAZ": "BL-12",
        "USANS": "BL-1A",
        "VENUS": "BL-10",
        "VISION": "BL-16B",
        "VULCAN": "BL-7",
    },
}


class AnalysisDataSelectorState(NeutronDataSelectorState):
    """Selection state for identifying datafiles."""

    allow_custom_directories: bool = Field(default=False)
    custom_directory: str = Field(default="", title="Custom Directory")

    @model_validator(mode="after")
    def validate_state(self) -> Self:
        valid_facilities = self.get_facilities()
        if self.facility and self.facility not in valid_facilities:
            warn(
                f"Facility '{self.facility}' could not be found. Valid options: {valid_facilities}",
                stacklevel=1,
            )

        valid_instruments = [instrument["name"] for instrument in self.get_instruments()]
        if self.instrument and self.facility != CUSTOM_DIRECTORIES_LABEL and self.instrument not in valid_instruments:
            warn(
                (
                    f"Instrument '{self.instrument}' could not be found in '{self.facility}'. "
                    f"Valid options: {valid_instruments}"
                ),
                stacklevel=1,
            )
        # Validating the experiment is expensive and will fail in our CI due to the filesystem not being mounted there.

        return self

    def get_facilities(self) -> List[str]:
        facilities = list(INSTRUMENT_IDS.keys())
        if self.allow_custom_directories:
            facilities.append(CUSTOM_DIRECTORIES_LABEL)
        return facilities

    def get_instruments(self) -> List[Dict[str, str]]:
        instruments = [
            {"id": id, "name": name, "title": f"{id}: {name}"}
            for name, id in INSTRUMENT_IDS.get(self.facility, {}).items()
        ]

        return natsorted(instruments, key=lambda x: x["name"])


class AnalysisDataSelectorModel(NeutronDataSelectorModel):
    """Analysis cluster filesystem backend for NeutronDataSelector."""

    def __init__(self, state: AnalysisDataSelectorState) -> None:
        super().__init__(state)
        self.state: AnalysisDataSelectorState = state

    def set_binding_parameters(self, **kwargs: Any) -> None:
        super().set_binding_parameters(**kwargs)

        if "allow_custom_directories" in kwargs:
            self.state.allow_custom_directories = kwargs["allow_custom_directories"]

    def get_custom_directory_path(self) -> Optional[Path]:
        # Don't expose the full file system
        if not self.state.custom_directory:
            return None

        return Path(self.state.custom_directory)

    def get_experiment_directory_path(self) -> Optional[Path]:
        if not self.state.experiment:
            return None

        return Path("/") / self.state.facility / self.get_instrument_dir() / self.state.experiment

    def get_instrument_dir(self) -> str:
        return INSTRUMENT_DIRS.get(self.state.facility, {}).get(self.state.instrument, "")

    def get_experiments(self) -> List[str]:
        experiments = []

        instrument_path = Path("/") / self.state.facility / self.get_instrument_dir()
        try:
            for dirname in os.listdir(instrument_path):
                if dirname.startswith("IPTS-") and os.access(instrument_path / dirname, mode=os.R_OK):
                    experiments.append(dirname)
        except OSError:
            pass

        return natsorted(experiments)

    def get_directories(self, base_path: Optional[Path] = None) -> List[Dict[str, Any]]:
        using_custom_directory = self.state.facility == CUSTOM_DIRECTORIES_LABEL
        if base_path:
            pass
        elif using_custom_directory:
            base_path = self.get_custom_directory_path()
        else:
            base_path = self.get_experiment_directory_path()

        if not base_path:
            return []

        return self.get_directories_from_path(base_path)

    def get_datafiles(self, *args: Any, **kwargs: Any) -> List[Dict[str, str]]:
        using_custom_directory = self.state.facility == CUSTOM_DIRECTORIES_LABEL
        if self.state.experiment:
            base_path = Path("/") / self.state.facility / self.get_instrument_dir() / self.state.experiment
        elif using_custom_directory and self.state.custom_directory:
            base_path = Path(self.state.custom_directory)
        else:
            return []

        return [{"path": datafile} for datafile in self.get_datafiles_from_path(base_path)]



"""Module for the Progress Tab."""

from typing import Tuple, Union

from trame.app import get_server
from trame.widgets import client
from trame.widgets import vuetify3 as vuetify
from trame_client.widgets import html

from nova.mvvm.trame_binding import TrameBinding
from nova.trame.view_model.execution_buttons import ExecutionButtonsViewModel


class ExecutionButtons:
    """Execution buttons class. Adds Run/Stop/Cancel/Download buttons to the view.

    This is intended to be used with the `nova-galaxy ToolRunner <https://nova-application-development.readthedocs.io/projects/nova-galaxy/en/latest/core_concepts/tool_runner.html>`__.
    """

    def __init__(self, id: str, stop_btn: Union[bool, Tuple] = False, download_btn: Union[bool, Tuple] = False) -> None:
        """Constructor for ExecutionButtons.

        For all parameters, tuples have a special syntax. See :ref:`TrameTuple <api_trame_tuple>` for a description of
        it.

        Parameters
        ----------
        id : str
            Component id. Should be used consistently with ToolRunner and other components. Note that this parameter
            does not support Trame bindings.
        stop_btn: Union[bool, Tuple]
            Display stop button.
        download_btn : Union[bool, Tuple]
            Display download button.

        Returns
        -------
        None
        """
        self.id = f"execution_{id}"

        self.server = get_server(None, client_type="vue3")
        binding = TrameBinding(self.server.state)
        self.ctrl = self.server.controller
        self.stop_btn = stop_btn
        self.download_btn = download_btn
        self.view_model = ExecutionButtonsViewModel(id, binding)
        self.view_model.buttons_state_bind.connect(self.id)
        self._download = client.JSEval(
            exec=(
                "async ($event) => {"
                " const blob = new window.Blob([$event], {type: 'application/zip'});"
                "  const url = window.URL.createObjectURL(blob);"
                "  const anchor = window.document.createElement('a');"
                "  anchor.setAttribute('href', url);"
                "  anchor.setAttribute('download', 'results.zip');"
                "  window.document.body.appendChild(anchor);"
                "  anchor.click();"
                "  window.document.body.removeChild(anchor);"
                "  setTimeout(() => window.URL.revokeObjectURL(url), 1000);"
                "}"
            )
        ).exec

        self.create_ui()

    def create_ui(self) -> None:
        with html.Div(classes="d-flex justify-center my-4 w-100"):
            vuetify.VBtn(
                "Run",
                disabled=(f"{self.id}.run_disabled",),
                prepend_icon="mdi-play",
                classes="mr-4",
                id=f"{self.id}_run",
                click=self.run,
            )
            if self.stop_btn:
                extra_params = {}
                if isinstance(self.stop_btn, tuple):
                    extra_params["v_if"] = self.stop_btn
                vuetify.VBtn(
                    "Stop",
                    disabled=(f"{self.id}.stop_disabled",),
                    loading=(f"{self.id}.stop_in_progress",),
                    classes="mr-4",
                    id=f"{self.id}_stop",
                    prepend_icon="mdi-stop",
                    click=self.stop,
                    **extra_params,
                )
            vuetify.VBtn(
                "Cancel",
                disabled=(f"{self.id}.cancel_disabled",),
                color="error",
                loading=(f"{self.id}.cancel_in_progress",),
                prepend_icon="mdi-cancel",
                classes="mr-4",
                id=f"{self.id}_cancel",
                click=self.cancel,
            )
            if self.download_btn:
                extra_params = {}
                if isinstance(self.download_btn, tuple):
                    extra_params["v_if"] = self.download_btn
                vuetify.VBtn(
                    "Download Results",
                    disabled=(f"{self.id}.download_disabled",),
                    loading=(f"{self.id}.download_in_progress",),
                    id=f"{self.id}.download",
                    click=self.download,
                    **extra_params,
                )

    async def download(self) -> None:
        content = await self.view_model.prepare_results()
        if content:
            self._download(content)

    async def run(self) -> None:
        await self.view_model.run()

    async def cancel(self) -> None:
        await self.view_model.cancel()

    async def stop(self) -> None:
        await self.view_model.stop()



"""Trame implementation of the GridLayout class."""

from typing import Any, Optional, Union
from warnings import warn

from trame.widgets import html
from trame_client.widgets.core import AbstractElement

from .utils import merge_styles


class GridLayout(html.Div):
    """Creates a grid with a specified number of columns."""

    def __init__(
        self,
        columns: int = 1,
        height: Optional[Union[int, str]] = None,
        width: Optional[Union[int, str]] = None,
        halign: Optional[str] = None,
        valign: Optional[str] = None,
        gap: Optional[Union[int, str]] = "0em",
        stretch: bool = False,
        **kwargs: Any,
    ) -> None:
        """Constructor for GridLayout.

        Parameters
        ----------
        columns : int
            The number of columns in the grid.
        height : optional[int | str]
            The height of this grid. If an integer is provided, it is interpreted as pixels. If a string is provided,
            the string is treated as a CSS value.
        width : optional[int | str]
            The width of this grid. If an integer is provided, it is interpreted as pixels. If a string is provided,
            the string is treated as a CSS value.
        halign : optional[str]
            The horizontal alignment of items in the grid. See `MDN
            <https://developer.mozilla.org/en-US/docs/Web/CSS/justify-items>`__ for available options.
        valign : optional[str]
            The vertical alignment of items in the grid. See `MDN
            <https://developer.mozilla.org/en-US/docs/Web/CSS/align-items>`__ for available options. Note that this
            parameter is ignored when stretch=True.
        gap : optional[str]
            The gap to place between items (works both horizontally and vertically). Can be any CSS gap value (e.g.
            "4px" or "0.25em"). Defaults to no gap between items.
        stretch : optional[bool]
            If True, then this layout component will stretch to attempt to fill the space of it's parent container.
            Defaults to False.
        kwargs : Any
            Additional keyword arguments to pass to html.Div.

        Returns
        -------
        None

        Examples
        --------
        Basic usage:

        .. literalinclude:: ../tests/gallery/views/app.py
            :start-after: setup grid
            :end-before: setup grid complete
            :dedent:

        Building a custom left-middle-right layout:

        .. literalinclude:: ../tests/test_layouts.py
            :start-after: setup complex layout example
            :end-before: setup complex layout example complete
            :dedent:
        """
        classes = kwargs.pop("classes", [])
        if isinstance(classes, list):
            classes = " ".join(classes)

        if stretch:
            if valign:
                warn("Ignoring valign parameter to GridLayout since stretch=True.", stacklevel=1)
            valign = "stretch"
            classes += " flex-1-1 overflow-y-auto"
        else:
            classes += " flex-0-1"

        v_show = kwargs.get("v_show", None)
        if v_show:
            classes = (f"{v_show} ? '{classes} d-grid' : '{classes}'",)
        else:
            classes += " d-grid"

        widget_style = self.get_root_styles(columns, height, width, halign, valign, gap)
        user_style = kwargs.pop("style", {})

        super().__init__(classes=classes, style=merge_styles(widget_style, user_style), **kwargs)

    def get_root_styles(
        self,
        columns: int,
        height: Optional[Union[int, str]],
        width: Optional[Union[int, str]],
        halign: Optional[str],
        valign: Optional[str],
        gap: Optional[Union[int, str]],
    ) -> dict[str, str]:
        height = f"{height}px" if isinstance(height, int) else height
        width = f"{width}px" if isinstance(width, int) else width
        gap = f"{gap}px" if isinstance(gap, int) else gap

        styles = {
            "grid-template-columns": f"repeat({columns}, 1fr)",
        }

        if height:
            styles["height"] = height
        if width:
            styles["width"] = width
        if halign:
            styles["justify-items"] = halign
        if valign:
            styles["align-items"] = valign
        if gap:
            styles["gap"] = gap

        return styles

    def get_row_style(self, row_span: int) -> str:
        return f"grid-row-end: span {row_span};"

    def get_column_style(self, column_span: int) -> str:
        return f"grid-column-end: span {column_span};"

    def add_child(self, child: Union[AbstractElement, str]) -> AbstractElement:
        """Add a child to the grid.

        Do not call this directly. Instead, use Trame's `with` syntax, which will call this method internally. This
        method is documented here as a reference for the span parameters.

        Parameters
        ----------
        child : `AbstractElement \
            <https://trame.readthedocs.io/en/latest/core.widget.html#trame_client.widgets.core.AbstractElement>`_ | str
            The child to add to the grid.
        row_span : int
            The number of rows this child should span.
        column_span : int
            The number of columns this child should span.

        Returns
        -------
        None

        Example
        -------
        .. literalinclude:: ../tests/gallery/views/app.py
            :start-after: grid row and column span example
            :end-before: grid row and column span example end
            :dedent:
        """
        if isinstance(child, str):
            child = html.Div(child)

        row_span = 1
        column_span = 1
        if "row_span" in child._py_attr:
            row_span = child._py_attr["row_span"]
        if "column_span" in child._py_attr:
            column_span = child._py_attr["column_span"]

        if "style" not in child._py_attr or child.style is None:
            child.style = ""
        child.style += f"; {self.get_row_style(row_span)} {self.get_column_style(column_span)}"

        if "classes" not in child._py_attr or child.classes is None:
            child.classes = ""
        child.classes += " d-grid-item"

        super().add_child(child)



"""Custom types for nova-trame."""

from typing import Any, NamedTuple

from typing_extensions import Self


class TrameTuple(NamedTuple):
    """Trame variable binding tuple.

    Trame allows you to set component parameters dynamically by passing a tuple to most parameters. This tuple requires
    the following syntax: (expression, initial_value). expression is typically a JavaScript variable, but it can be any
    JavaScript expression that evaluates to a boolean value. initial_value should only used if your expression is a
    JavaScript variable.
    """

    expression: str
    initial_value: Any = None

    @classmethod
    def create(cls, expression: Any, initial_value: Any = None) -> Self:
        """Turns a generic expression into a tuple suitable for binding a Trame component.

        This is primarily intended for internal usage and makes aggressive assumptions about how to convert Python types
        to JavaScript expressions. If you are creating your own components with parameters that accept built-in types
        and Trame binding tuples, then it may prove useful to you either as is or subclassed.

        Parameters
        ----------
        expression : Any
            The content to turn into a JavaScript expression. If a string-type is provided, it is passed to Trame
            without modification.
        initial_value: Any, optional
            The initial value to assign to expression. If set, expression must be a JavaScript variable.
        """
        if (
            isinstance(expression, tuple)
            and len(expression) > 0
            and len(expression) < 3
            and isinstance(expression[0], str)
        ):
            # This looks like a Trame-style tuple, we should treat it accordingly.
            if len(expression) == 2:
                initial_value = expression[1]
            expression = expression[0]

        if isinstance(expression, (bool, complex, float, int)):
            if expression:
                return cls("true")
            else:
                return cls("false")

        return cls(str(expression), initial_value)



"""View Implementation for InputField."""

import logging
import os
import re
from enum import Enum
from inspect import isclass
from typing import Any, Dict, Tuple, Union

from trame.app import get_server
from trame.widgets import client
from trame.widgets import vuetify3 as vuetify
from trame_client.widgets.core import AbstractElement
from trame_server.controller import Controller
from trame_server.state import State

from nova.mvvm.pydantic_utils import ERROR_FIELD_NAME, get_field_info
from nova.trame._internal.utils import set_state_param

logger = logging.getLogger(__name__)


def parse_v_model(v_model: Union[str, Tuple, None]) -> Tuple[str, str]:
    if isinstance(v_model, str):
        field = v_model
    elif isinstance(v_model, tuple):
        field = v_model[0]
    else:
        field = ""
    return field, field.split(".")[0]


class InputField(AbstractElement):
    """Factory class for generating Vuetify input components."""

    next_id = 0

    @staticmethod
    def create_boilerplate_properties(
        v_model: Union[str, Tuple, None],
        field_type: str,
        debounce: Union[int, Tuple],
        throttle: Union[int, Tuple],
    ) -> dict:
        if debounce == -1:
            debounce = int(os.environ.get("NOVA_TRAME_DEFAULT_DEBOUNCE", 0))
        if throttle == -1:
            throttle = int(os.environ.get("NOVA_TRAME_DEFAULT_THROTTLE", 0))

        if not v_model:
            return {}
        field, object_name_in_state = parse_v_model(v_model)
        field_info = None
        try:
            field_name = ".".join(field.split(".")[1:])
            if "[" in field_name:
                index_field_name = re.sub(r"\[.*?\]", "[0]", field_name)
                field_info = get_field_info(f"{object_name_in_state}.{index_field_name}")
                if "[" in field_name and "[index]" not in field_name:
                    field_info = None
                    logger.warning(
                        f"{field}: validation ignored. We currently only "
                        f"support single loop with index variable that should be called 'index'"
                    )
            else:
                field_info = get_field_info(field)
        except Exception as _:
            pass
        label = ""
        help_dict: dict = {}
        placeholder = None
        if field_info:
            label = field_info.title
            if field_info.examples and len(field_info.examples) > 0:
                placeholder = field_info.examples[0]
            help_dict = {"hint": field_info.description, "placeholder": placeholder}

        args: Dict[str, Any] = {}
        if v_model:
            args |= {"v_model": v_model, "label": label, "help": help_dict}
            if field_info:
                args |= {
                    "rules": (f"[(v) => trigger('validate_pydantic_field', ['{field}', v, index])]",),
                }

                if (
                    field_type in ["autocomplete", "combobox", "select"]
                    and isclass(field_info.annotation)
                    and issubclass(field_info.annotation, Enum)
                ):
                    args |= {"items": str([option.value for option in field_info.annotation])}

            if debounce and throttle:
                raise ValueError("debounce and throttle cannot be used together")

            server = get_server(None, client_type="vue3")
            if debounce:
                if isinstance(debounce, tuple):
                    debounce_field = debounce[0]
                    set_state_param(server.state, debounce)
                else:
                    debounce_field = f"nova__debounce_{InputField.next_id}"
                    InputField.next_id += 1
                    set_state_param(server.state, debounce_field, debounce)

                args |= {
                    "update_modelValue": (
                        "window.delay_manager.debounce("
                        f"  '{field}',"
                        f"  () => flushState('{object_name_in_state}'),"
                        f"  {debounce_field}"
                        ")"
                    )
                }
            elif throttle:
                if isinstance(throttle, tuple):
                    throttle_field = throttle[0]
                    set_state_param(server.state, throttle)
                else:
                    throttle_field = f"nova__throttle_{InputField.next_id}"
                    InputField.next_id += 1
                    set_state_param(server.state, throttle_field, throttle)

                args |= {
                    "update_modelValue": (
                        "window.delay_manager.throttle("
                        f"  '{field}',"
                        f"  () => flushState('{object_name_in_state}'),"
                        f"  {throttle_field}"
                        ")"
                    )
                }
            else:
                args |= {"update_modelValue": f"flushState('{object_name_in_state}')"}
        return args

    def __new__(
        cls,
        v_model: Union[str, Tuple, None] = None,
        required: bool = False,
        debounce: Union[int, Tuple] = -1,
        throttle: Union[int, Tuple] = -1,
        type: str = "text",
        **kwargs: Any,
    ) -> AbstractElement:
        """Constructor for InputField.

        For all parameters, tuples have a special syntax. See :ref:`TrameTuple <api_trame_tuple>` for a description of
        it.

        Parameters
        ----------
        v_model : Union[str, Tuple], optional
            The v-model for this component. If this references a Pydantic configuration variable, then this component
            will attempt to load a label, hint, and validation rules from the configuration for you automatically.
        required : bool, optional
            If true, the input will be visually marked as required and a required rule will be added to the end of the
            rules list. This parameter will be removed in the future. Please use Pydantic to enforce validation of
            required fields.
        debounce : Union[int, Tuple], optional
            Number of milliseconds to wait after the last user interaction with this field before attempting to update
            the Trame state. If set to 0, then no debouncing will occur. If set to -1, then the environment variable
            `NOVA_TRAME_DEFAULT_DEBOUNCE` will be used to set this (defaults to 0). See the `Lodash Docs
            <https://lodash.com/docs/4.17.15#debounce>`__ for details.
        throttle : Union[int, Tuple], optional
            Number of milliseconds to wait between updates to the Trame state when the user is interacting with this
            field. If set to 0, then no throttling will occur. If set to -1, then the environment variable
            `NOVA_TRAME_DEFAULT_THROTTLE` will be used to set this (defaults to 0). See the `Lodash Docs
            <https://lodash.com/docs/4.17.15#throttle>`__ for details.
        type : str
            The type of input to create. This can be any of the following:

            - autocomplete - Produces a dropdown menu that supports autocompletion as the user types. Items can be \
                automatically populated (see select option for details).
            - autoscroll - Produces a textarea that automatically scrolls to the bottom as new content is added.
            - checkbox
            - combobox - Produces a dropdown menu that supports autocompletion as the user types and allows users to \
                add new items. Items can be automatically populated (see select option for details).
            - file
            - input
            - number
            - otp
            - radio - Produces a radio button group. Note that this accepts an additional parameter items that expects \
                a list of dictionaries with the following format: { title: 'Item 1', value: 'item_1' }.
            - range-slider
            - select - Produces a dropdown menu. This menu can have items automatically populated if the v_model is \
                connected to a Pydantic field that uses an Enum type. Otherwise, you must specify the items parameter \
                to `InputField`.

                .. literalinclude:: ../tests/test_input_field.py
                    :start-after: items autopopulation start
                    :end-before: items autopopulation end
                    :dedent:

                .. code-block:: python

                    InputField(v_model="dropdown.enum_field", type="select")

            - slider
            - switch
            - textarea

            Any other value will produce a text field with your type used as an HTML input type attribute. Note that
            parameter does not support binding since swapping field types dynamically produces a confusing user
            experience.
        **kwargs
            All other arguments will be passed to the underlying
            `Trame Vuetify component <https://trame.readthedocs.io/en/latest/trame.widgets.vuetify3.html>`_.
            The following example would set the auto_grow and label attributes on
            `VTextarea <https://trame.readthedocs.io/en/latest/trame.widgets.vuetify3.html#trame.widgets.vuetify3.VTextarea>`_:

            .. literalinclude:: ../tests/gallery/views/app.py
                :start-after: InputField kwargs example start
                :end-before: InputField kwargs example end
                :dedent:

        Returns
        -------
        `trame_client.widgets.core.AbstractElement <https://trame.readthedocs.io/en/latest/core.widget.html#trame_client.widgets.core.AbstractElement>`_
            The Vuetify input component.
        """
        server = get_server(None, client_type="vue3")

        kwargs = {
            **cls.create_boilerplate_properties(v_model, type, debounce, throttle),
            **kwargs,
        }

        if "__events" not in kwargs or kwargs["__events"] is None:
            kwargs["__events"] = []

        # This must be present before each input is created or change events won't be triggered.
        if isinstance(kwargs["__events"], list):
            if "blur" not in kwargs["__events"]:
                kwargs["__events"].append("blur")
            if "change" not in kwargs["__events"]:
                kwargs["__events"].append("change")
            if "scroll" not in kwargs["__events"]:
                kwargs["__events"].append("scroll")

        match type:
            case "autocomplete":
                input = vuetify.VAutocomplete(**kwargs)
            case "autoscroll":
                input = vuetify.VTextarea(**kwargs)
                cls._setup_autoscroll(server.state, input)
            case "checkbox":
                input = vuetify.VCheckbox(**kwargs)
            case "combobox":
                input = vuetify.VCombobox(**kwargs)
            case "file":
                input = vuetify.VFileInput(**kwargs)
            case "input":
                input = vuetify.VInput(**kwargs)
            case "number":
                input = vuetify.VNumberInput(**kwargs)
            case "otp":
                input = vuetify.VOtpInput(**kwargs)
            case "radio":
                items = kwargs.pop("items", None)
                if isinstance(items, tuple):
                    items = items[0]
                with vuetify.VRadioGroup(**kwargs) as input:
                    vuetify.VRadio(v_for=f"item in {items}", label=("item.title",), value=("item.value",))
            case "range-slider":
                input = vuetify.VRangeSlider(**kwargs)
            case "select":
                items = kwargs.pop("items", None)
                if isinstance(items, str):
                    items = (items,)
                input = vuetify.VSelect(items=items, **kwargs)
            case "slider":
                input = vuetify.VSlider(**kwargs)
            case "switch":
                input = vuetify.VSwitch(**kwargs)
            case "textarea":
                input = vuetify.VTextarea(**kwargs)
            case _:
                input = vuetify.VTextField(type=type, **kwargs)

        cls._setup_ref(input)
        cls._setup_help(input, **kwargs)

        cls._check_rules(input)
        if required:
            cls._setup_required_label(input)
            cls._setup_required_rule(input)

        cls._setup_event_listeners(server.controller, input)

        return input

    @staticmethod
    def _check_rules(input: AbstractElement) -> None:
        if "rules" in input._py_attr and input.rules and not isinstance(input.rules, tuple):
            raise ValueError(f"Rules for '{input.label}' must be a tuple")

    @staticmethod
    def _setup_autoscroll(state: State, input: AbstractElement) -> None:
        if input.v_model:
            if "id" not in input._py_attr or input.id is None:
                input.id = f"nova__{input._id}"
            input.scroll = f"window.nova__autoscroll('{input.id}');"

            with state:
                if state["nova_scroll_position"] is None:
                    state["nova_scroll_position"] = {}
                state.nova_scroll_position[input.id] = 0

            autoscroll = client.JSEval(
                exec=(
                    "if (window.nova__autoscroll !== undefined) {"
                    # If the autoscroll function already exists, call it.
                    "  window.nova__autoscroll($event);"
                    "} else {"
                    # Save the JS so it can be called from outside of this script (ie during a scroll event).
                    "  window.nova__autoscroll = function(id) {"
                    # Get the element in the browser by ID
                    "    const element = window.document.querySelector(`#${id}`);"
                    # If the user is at the bottom of the textarea, then we should autoscroll.
                    "    if (element.scrollTop === window.trame.state.state.nova_scroll_position[id]) {"
                    # Scroll to the bottom
                    "      element.scrollTop = element.scrollHeight;"
                    # Save the new scroll position
                    "      window.trame.state.state.nova_scroll_position[id] = element.scrollTop;"
                    "      flushState('nova_scroll_position');"
                    # If the user has scrolled back to the bottom, then we should reenable scrolling.
                    "    } else if (element.scrollTop === element.scrollHeight - element.clientHeight) {"
                    # Save the new scroll position
                    "      window.trame.state.state.nova_scroll_position[id] = element.scrollTop;"
                    "      flushState('nova_scroll_position');"
                    "    }"
                    "  };"
                    "  window.nova__autoscroll($event);"
                    "}"
                )
            ).exec

            @state.change(input.v_model.split(".")[0])
            def _autoscroll(**kwargs: Any) -> None:
                autoscroll(input.id)

    @staticmethod
    def _setup_help(_input: AbstractElement, **kwargs: Any) -> None:
        help = kwargs.get("help", None)
        if help and isinstance(help, dict):
            _input.placeholder = help.get("placeholder", None)
            _input.title = help.get("hint", None)

    @staticmethod
    def _setup_required_label(input: AbstractElement) -> None:
        if input.label:
            input.label = f"{input.label}*"
        else:
            input.label = "*"

    @staticmethod
    def _setup_ref(input: AbstractElement) -> None:
        if "ref" not in input._py_attr or input.ref is None:
            input.ref = f"nova__{input._id}"

        if "v_model" in input._py_attr and input.v_model is not None:
            with input:
                _, object_name_in_state = parse_v_model(input.v_model)
                client.ClientTriggers(
                    mounted=(
                        f"if (window.trame.state.state?.{object_name_in_state}?.{ERROR_FIELD_NAME}?.length > 0) {{"
                        f"    window.setTimeout(() => {{ window.trame.refs['{input.ref}'].validate(); }}, 100);"
                        "}"
                    )
                )

    @staticmethod
    def _setup_required_rule(input: AbstractElement) -> None:
        # The rule needs to check that 1. the input has been touched by the user, and 2. the input is not empty.
        required_rule = (
            f"(value) => (!window.trame.refs['{input.ref}'].touched || value?.length > 0) || 'Field is required'"
        )
        if "rules" in input._py_attr and input.rules:
            # Existing rules will be in format ("[rule1, rule2]",) and we need to append to this list
            rule_end_index = input.rules[0].rindex("]")
            input.rules = (f"{input.rules[0][:rule_end_index]}, {required_rule}{input.rules[0][rule_end_index:]}",)
        else:
            input.rules = (f"[{required_rule}]",)

    @staticmethod
    def _setup_event_listeners(ctrl: Controller, input: AbstractElement) -> None:
        base_handler = None
        if "change" in input._py_attr and input.change is not None:
            base_handler = input.change

        # Iterate over all saved refs and perform validation if there is a value that can be validated.
        change_handler = (
            "Object.values(window.trame.refs).map("
            "  (ref) => ref && typeof ref.validate === 'function' && ref.value ? ref.validate() : null"
            ");"
        )

        # We need to coerce the developer's change handler, which could be a string, callable, or tuple containing a
        # callable, to a single string to be compatible with our change handler.
        if callable(base_handler):
            base_handler = (base_handler,)
        if isinstance(base_handler, tuple):

            @ctrl.trigger(f"{input.ref}__trigger")
            def _(*args: str, **kwargs: Any) -> None:
                base_handler[0](*args, **kwargs)

            change_handler = (
                "trigger("
                f"'{input.ref}__trigger', "
                f"{base_handler[1] if len(base_handler) > 1 else []}, "
                f"{base_handler[2] if len(base_handler) > 2 else {} }"
                f"); {change_handler}"
            )  # Call the developer's provided change method via a trigger, then call ours.
        elif isinstance(base_handler, str):
            # Call the developer's provided change JS expression, then call ours.
            change_handler = f"{base_handler}; {change_handler}"

        # The user touched the input, so we can enable the required rule.
        input.blur = f"window.trame.refs['{input.ref}'].touched = true"
        input.change = change_handler



"""Module for the JobProgress ViewModel."""

import json
from typing import Any, Dict, Tuple

import blinker
from pydantic import BaseModel

from nova.common.job import WorkState
from nova.common.signals import Signal, get_signal_id
from nova.mvvm.interface import BindingInterface


def details_from_state(state: WorkState, details: Dict[str, Any]) -> Tuple[str, str]:
    work_state_map = {
        WorkState.NOT_STARTED: "job not started",
        WorkState.UPLOADING_DATA: "uploading data",
        WorkState.QUEUED: "job is queued",
        WorkState.RUNNING: "job is running",
        WorkState.FINISHED: "job finished",
        WorkState.ERROR: "job produced an error",
        WorkState.DELETED: "job deleted",
        WorkState.CANCELED: "job canceled",
        WorkState.STOPPING: "stopping job",
        WorkState.CANCELING: "canceling job",
    }
    if state in work_state_map:
        state_str = work_state_map[state]
    else:
        state_str = state.value

    full_details_dict = json.dumps(details.get("original_dict", {}), indent=4)
    return state_str, full_details_dict


class ProgressState(BaseModel):
    """Class that manages progress bars states."""

    progress: str = ""
    details: str = ""
    full_details: str = ""
    show_progress: bool = False
    show_full_details: bool = False
    show_failed: bool = False
    show_ok: bool = False

    def update_from_workstate(self, state: WorkState, details: Dict[str, Any]) -> None:
        progress = "0"
        match state:
            case WorkState.UPLOADING_DATA:
                progress = "10"
            case WorkState.QUEUED:
                progress = "20"
            case WorkState.RUNNING:
                progress = "50"

        self.show_progress = state in [
            WorkState.UPLOADING_DATA,
            WorkState.QUEUED,
            WorkState.RUNNING,
            WorkState.CANCELING,
            WorkState.STOPPING,
        ]
        self.show_failed = state == WorkState.ERROR
        self.show_ok = state == WorkState.FINISHED
        self.progress = progress
        self.details, self.full_details = details_from_state(state, details)
        self.show_full_details = self.full_details != "{}"


class ProgressBarViewModel:
    """A viewmodel responsible for progress bar."""

    def __init__(self, id: str, binding: BindingInterface):
        self.progress_state = ProgressState()
        self.progress_state_bind = binding.new_bind(self.progress_state)
        self.progress_signal = blinker.signal(get_signal_id(id, Signal.PROGRESS))
        self.progress_signal.connect(self.update_state, weak=False)

    async def update_state(self, _sender: Any, state: WorkState, details: Dict[str, Any]) -> None:
        self.progress_state.update_from_workstate(state, details)
        self.progress_state_bind.update_in_view(self.progress_state)



"""Unit tests for Interactive2DPlot."""

from selenium.webdriver import ActionChains, Firefox
from selenium.webdriver.common.by import By

from nova.trame.view.components.visualization import Interactive2DPlot


def test_interactive_2d_plot() -> None:
    # [setup 2d plot]
    from altair import Chart, selection_interval
    from vega_datasets import data

    brush = selection_interval(name="brush")
    chart = (
        Chart(data.cars())
        .mark_circle()
        .encode(x="Horsepower:Q", y="Miles_per_Gallon:Q", color="Origin:N")
        .add_params(brush)
    )
    plot = Interactive2DPlot(figure=chart)
    # [setup 2d plot complete]
    assert plot._figure == chart


def test_missing_figure() -> None:
    plot = Interactive2DPlot()
    assert plot._figure is None


def test_state_synchronization(driver: Firefox) -> None:
    plot = driver.find_element(By.ID, "interactive-plot")
    ActionChains(driver).drag_and_drop_by_offset(plot, 10, 5).perform()

    # Look for the interval in the state. I want to avoid using the exact ref since it's pseudo-random
    # (e.g. nova__vega_203) here.
    trame_state = driver.execute_script("return window.trame.state.state")
    for key, value in trame_state.items():
        if isinstance(value, dict):
            for sub_key in value:
                if sub_key == "interval" and key.startswith("nova__vega"):
                    interval = value[sub_key]

    # We could check exact values, but that seems more sensitive to changes in the gallery.
    assert "Horsepower" in interval and isinstance(interval["Horsepower"], list)



"""View implementation for RemoteFileInput."""

from functools import partial
from tempfile import NamedTemporaryFile
from typing import Any, Iterable, Optional, Tuple, Union, cast

from trame.app import get_server
from trame.widgets import client, html
from trame.widgets import vuetify3 as vuetify
from trame_client.widgets.core import AbstractElement
from trame_server.core import State

from nova.mvvm._internal.utils import rgetdictvalue
from nova.mvvm.trame_binding import TrameBinding
from nova.trame._internal.utils import get_state_name, get_state_param, set_state_param
from nova.trame.model.remote_file_input import RemoteFileInputModel
from nova.trame.view.layouts import VBoxLayout
from nova.trame.view_model.remote_file_input import RemoteFileInputViewModel

from .input_field import InputField


class RemoteFileInput:
    """Generates a file selection dialog for picking files off of the server.

    You cannot use typical Trame :code:`with` syntax to add children to this.
    """

    def __init__(
        self,
        v_model: Union[str, Tuple],
        allow_files: Union[bool, Tuple] = True,
        allow_folders: Union[bool, Tuple] = False,
        base_paths: Union[Iterable[str], str, None] = None,
        dialog_props: Optional[dict[str, Any]] = None,
        extensions: Union[Iterable[str], str, None] = None,
        input_props: Optional[dict[str, Any]] = None,
        return_contents: Union[bool, Tuple] = False,
        use_bytes: Union[bool, Tuple] = False,
    ) -> None:
        """Constructor for RemoteFileInput.

        For all parameters, tuples have a special syntax. See :ref:`TrameTuple <api_trame_tuple>` for a description of
        it.

        Parameters
        ----------
        v_model : Union[str, Tuple]
            The v-model for this component. If this references a Pydantic configuration variable, then this component
            will attempt to load a label, hint, and validation rules from the configuration for you automatically.
        allow_files : Union[bool, Tuple], optional
            If true, the user can save a file selection.
        allow_folders : Union[bool, Tuple], optional
            If true, the user can save a folder selection.
        base_paths : Union[Iterable[str], str], optional
            Only files under these paths will be shown. Typical Trame binding syntax doesn't work here as tuples are
            interpreted as literal extensions to filter with. Instead, you can pass a string with a JavaScript
            expression to bind this parameter.
        dialog_props : Dict[str, typing.Any], optional
            Props to be passed to VDialog.
        extensions : Union[Iterable[str], str], optional
            Only files with these extensions will be shown by default. The user can still choose to view all files.
            Typical Trame binding syntax doesn't work here as tuples are interpreted as literal extensions to filter
            with. Instead, you can pass a string with a JavaScript expression to bind this parameter.
        input_props : Dict[str, typing.Any], optional
            Props to be passed to InputField.
        return_contents : Union[bool, Tuple], optional
            If true, then the v_model will contain the contents of the file. If false, then the v_model will contain the
            path of the file. Defaults to false.
        use_bytes : Union[bool, Tuple], optional
            If true, then the file contents will be treated as bytestreams when calling decode_file.

        Returns
        -------
        None
        """
        self.server = get_server(None, client_type="vue3")

        self.v_model = v_model
        self.allow_files = allow_files
        self.allow_folders = allow_folders
        self.base_paths = base_paths if base_paths else ["/"]
        self.dialog_props = dict(dialog_props) if dialog_props else {}
        self.extensions = extensions if extensions else []
        self.input_props = dict(input_props) if input_props else {}
        self.return_contents = return_contents
        self.use_bytes = use_bytes

        if "__events" not in self.input_props:
            self.input_props["__events"] = []
        self.input_props["__events"].append("change")

        if "width" not in self.dialog_props:
            self.dialog_props["width"] = 600

        self._create_model()
        self._create_viewmodel()
        self._setup_bindings()

        self.create_ui()

    @property
    def state(self) -> State:
        return self.server.state

    def create_ui(self) -> None:
        with cast(
            AbstractElement,
            InputField(
                v_model=self.v_model,
                change=(self.vm.select_file, "[$event.target.value]"),
                **self.input_props,
            ),
        ) as input:
            if isinstance(input.classes, str):
                input.classes += " nova-remote-file-input"
            else:
                input.classes = "nova-remote-file-input"
            self.vm.init_view()

            with vuetify.Template(v_slot_append=True):
                with vuetify.VBtn(icon=True, size="small", click=self.vm.open_dialog):
                    vuetify.VIcon("mdi-folder-open")

                    with vuetify.VDialog(
                        v_model=self.vm.get_dialog_state_name(), activator="parent", height="100vh", **self.dialog_props
                    ):
                        with vuetify.VCard(classes="pa-4"):
                            with VBoxLayout():
                                vuetify.VTextField(
                                    v_model=self.vm.get_filter_state_name(),
                                    classes="mb-4 px-4",
                                    label=input.label,
                                    variant="outlined",
                                    update_modelValue=(self.vm.filter_paths, "[$event]"),
                                )

                            if self.allow_files and self.extensions:
                                with html.Div(v_if=(f"{self.vm.get_showing_all_state_name()}",)):
                                    vuetify.VListSubheader("All Available Files")
                                    vuetify.VBtn(
                                        "Don't show all",
                                        classes="mb-4",
                                        size="small",
                                        click=self.vm.toggle_showing_all_files,
                                    )
                                with html.Div(v_else=True):
                                    vuetify.VListSubheader(
                                        "Available Files with Extensions: "
                                        + ", ".join(
                                            get_state_param(self.state, (self.extensions,))
                                            if isinstance(self.extensions, str)
                                            else self.extensions
                                        )
                                    )
                                    vuetify.VBtn(
                                        "Show all",
                                        classes="mb-4",
                                        size="small",
                                        click=self.vm.toggle_showing_all_files,
                                    )
                            elif self.allow_files:
                                vuetify.VListSubheader("Available Files")
                            else:
                                vuetify.VListSubheader("Available Folders")

                            with vuetify.VList(classes="mb-4"):
                                self.vm.populate_file_list()

                                vuetify.VListItem(
                                    "{{ file.path }}",
                                    v_for=f"file, index in {self.vm.get_file_list_state_name()}",
                                    classes=(
                                        f"index < {self.vm.get_file_list_state_name()}.length - 1 "
                                        "? 'border-b-thin' "
                                        ": ''",
                                    ),
                                    prepend_icon=("file.directory ? 'mdi-folder' : 'mdi-file'",),
                                    click=(self.vm.select_file, "[file]"),
                                )
                                html.P(
                                    (
                                        "No files could be found, either because none exist or you lack permission to "
                                        "read this directory. Select .. to return to the previous directory."
                                    ),
                                    v_if=(
                                        f"{self.vm.get_file_list_state_name()}.length < 2 && "
                                        f"{self.vm.get_file_list_state_name()}[0].path === '..'"
                                    ),
                                    classes="pa-4",
                                )

                            vuetify.VSpacer()

                            with html.Div(classes="text-center"):
                                vuetify.VBtn(
                                    "OK",
                                    classes="mr-4",
                                    disabled=(f"!{self.vm.get_valid_selection_state_name()}",),
                                    click=self.vm.close_dialog,
                                )
                                vuetify.VBtn(
                                    "Cancel",
                                    color="lightgrey",
                                    click=partial(self.vm.close_dialog, cancel=True),
                                )

    def _create_model(self) -> None:
        self.model = RemoteFileInputModel()

    def _create_viewmodel(self) -> None:
        binding = TrameBinding(self.state)

        if isinstance(self.v_model, tuple):
            model_name = self.v_model[0]
        else:
            model_name = self.v_model

        self.set_v_model = client.JSEval(
            exec=f"{model_name} = $event; flushState('{model_name.split('.')[0].split('[')[0]}');"
        ).exec

        self.vm = RemoteFileInputViewModel(self.model, binding)

        self.vm.dialog_bind.connect(self.vm.get_dialog_state_name())
        self.vm.file_list_bind.connect(self.vm.get_file_list_state_name())
        self.vm.filter_bind.connect(self.vm.get_filter_state_name())
        self.vm.on_close_bind.connect(client.JSEval(exec=f"{self.vm.get_dialog_state_name()} = false;").exec)
        self.vm.showing_all_bind.connect(self.vm.get_showing_all_state_name())
        self.vm.valid_selection_bind.connect(self.vm.get_valid_selection_state_name())

    # This method sets up Trame state change listeners for each binding parameter that can be changed directly by this
    # component. This allows us to communicate the changes to the developer's bindings without requiring our own. We
    # don't want bindings in the internal implementation as our callbacks could compete with the developer's.
    def _setup_bindings(self) -> None:
        # If the bindings were given initial values, write these to the state.
        self._last_allow_files = set_state_param(self.state, self.allow_files)
        self._last_allow_folders = set_state_param(self.state, self.allow_folders)
        self._last_base_paths = (
            get_state_param(self.state, (self.base_paths,)) if isinstance(self.base_paths, str) else self.base_paths
        )
        self._last_extensions = (
            get_state_param(self.state, (self.extensions,)) if isinstance(self.extensions, str) else self.extensions
        )
        self._last_return_contents = set_state_param(self.state, self.return_contents)

        # Now we need to propagate the state to this component's view model.
        self.vm.set_binding_parameters(
            allow_files=self.allow_files,
            allow_folders=self.allow_folders,
            base_paths=self._last_base_paths,
            extensions=self._last_extensions,
        )
        self._setup_update_binding(self._last_return_contents)

        # Now we set up the change listeners for all bound parameters. These are responsible for updating the component
        # when other portions of the application manipulate these parameters.
        if isinstance(self.allow_files, tuple):

            @self.state.change(get_state_name(self.allow_files[0]))
            def on_allow_files_change(**kwargs: Any) -> None:
                if isinstance(self.allow_files, bool):
                    return
                allow_files = rgetdictvalue(kwargs, self.allow_files[0])
                if allow_files != self._last_allow_files:
                    self.vm.set_binding_parameters(
                        allow_files=set_state_param(self.state, self.allow_files, allow_files)
                    )

        if isinstance(self.allow_folders, tuple):

            @self.state.change(get_state_name(self.allow_folders[0]))
            def on_allow_folders_change(**kwargs: Any) -> None:
                if isinstance(self.allow_folders, bool):
                    return
                allow_folders = rgetdictvalue(kwargs, self.allow_folders[0])
                if allow_folders != self._last_allow_folders:
                    self.vm.set_binding_parameters(
                        allow_folders=set_state_param(self.state, self.allow_folders, allow_folders)
                    )

        if isinstance(self.base_paths, str):

            @self.state.change(get_state_name(self.base_paths))
            def on_base_paths_change(**kwargs: Any) -> None:
                if isinstance(self.base_paths, bool):
                    return
                base_paths = rgetdictvalue(kwargs, self.base_paths)
                if base_paths != self._last_base_paths:
                    self.vm.set_binding_parameters(base_paths=set_state_param(self.state, self.base_paths, base_paths))

        if isinstance(self.extensions, str):

            @self.state.change(get_state_name(self.extensions))
            def on_extensions_change(**kwargs: Any) -> None:
                if isinstance(self.extensions, bool):
                    return
                extensions = rgetdictvalue(kwargs, self.extensions)
                if extensions != self._last_extensions:
                    self.vm.set_binding_parameters(extensions=set_state_param(self.state, self.extensions, extensions))

        if isinstance(self.return_contents, tuple):

            @self.state.change(get_state_name(self.return_contents[0]))
            def on_return_contents_change(**kwargs: Any) -> None:
                if isinstance(self.return_contents, bool):
                    return
                return_contents = rgetdictvalue(kwargs, self.return_contents[0])
                if return_contents != self._last_return_contents:
                    self._setup_update_binding(return_contents)

    def _setup_update_binding(self, read_file: bool) -> None:
        self.vm.reset_update_binding()
        if read_file:
            self.vm.on_update_bind.connect(self.read_file)
        else:
            self.vm.on_update_bind.connect(self.set_v_model)

    def read_file(self, file_path: str) -> None:
        with open(file_path, mode="rb") as file:
            self.decode_file(file.read())

    def decode_file(self, bytestream: bytes, set_contents: bool = False) -> None:
        use_bytes = get_state_param(self.state, self.use_bytes)

        decoded_content = bytestream.decode("latin1")
        if set_contents:
            self.set_v_model(decoded_content)
        else:
            if use_bytes:
                with NamedTemporaryFile(mode="wb", delete=False) as temp_file:
                    temp_file.write(bytestream)
                    temp_file.flush()
                    self.set_v_model(temp_file.name)
            else:
                with NamedTemporaryFile(mode="w", delete=False, encoding="utf-8") as temp_file:
                    temp_file.write(decoded_content)
                    temp_file.flush()
                    self.set_v_model(temp_file.name)

    def select_file(self, value: str) -> None:
        """Programmatically set the v_model value."""
        self.vm.select_file(value)

    def open_dialog(self) -> None:
        """Programmatically opens the dialog for selecting a file."""
        self.vm.open_dialog()
